from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *

from libc.stdlib cimport abs

import collections, array, struct, itertools
import CGAT.Experiment as E

def merge_pairs(AlignmentFile input_samfile,
                outfile,
                min_insert_size = 0,
                max_insert_size = 400,
                bed_format = None ):
    '''merge paired ended data.

    For speed reasons, the aligned region is only approximated using
    the read length. Indels within reads are ignored. Thus bed coordinates
    might be a few residues off.

    The strand is always set to '+'.

    Pairs with a maximum insert size larger than *max_insert_size* are removed.

    If `bed_format` is a number, only the first x columns will be output.
    '''

    cdef int ninput = 0
    cdef int nremoved_unmapped = 0
    cdef int nremoved_insert = 0
    cdef int nremoved_contig = 0
    cdef int nremoved_unpaired = 0
    cdef int nremoved_coordinate = 0
    cdef int nremoved_take_only_second = 0
    cdef int noutput = 0
    cdef int flag
    cdef int isize

    cdef AlignedSegment read
    cdef int c_max_insert_size = max_insert_size
    cdef int c_min_insert_size = min_insert_size
    cdef int start, end, xstart, xend
    cdef int take_columns = 6

    # point to array of contig lengths
    cdef uint32_t *contig_sizes = input_samfile.header.target_len
    
    if bed_format != None:
        if bed_format < 3 or bed_format > 6: 
            raise ValueError("a bed file must have at least 3 and at most 6 columns")
        take_columns = bed_format

    for read in input_samfile:
        ninput += 1

        flag = read._delegate.core.flag 
        # remove unmapped reads
        if flag & 4:
            nremoved_unmapped += 1
            continue

        if read.pos < read.mpos:
            # lower coordinate than mate, ignore
            nremoved_coordinate += 1
            continue
        elif read.pos == read.mpos and flag & 64:
            # disambiguate, ignore first in pair
            nremoved_take_only_second += 1
            continue
        else:
            # taking the downstream pair allows to incl
            xstart = read.next_reference_start
            xend = read.reference_end
            if xstart < xend:
                start, end = xstart, xend
            else:
                start, end = xend, xstart

        # remove unpaired
        if not flag & 2:
            nremoved_unpaired += 1
            continue
            
        if read.tid != read.mrnm:
            nremoved_contig += 1
            continue

        # isize can be negative - depending on the pair orientation
        isize = abs(read.isize)
        if (c_max_insert_size and isize > c_max_insert_size) or \
           (c_min_insert_size and isize < c_min_insert_size):
            nremoved_insert += 1
            continue

        # truncate at contig end - overhanging reads might cause problems with chrM
        if end > contig_sizes[read.mrnm]:
            end = contig_sizes[read.mrnm]

        # count output pair as two so that it squares with ninput
        noutput += 2

        if take_columns == 3:
            outfile.write("%s\t%i\t%i\n" %
                          (input_samfile.getrname(read.tid),
                           start, end))
        elif take_columns == 4:
            outfile.write("%s\t%i\t%i\t%s\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname))
        elif take_columns == 5:
            outfile.write("%s\t%i\t%i\t%s\t%i\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname,
                           read.mapq,
                       ))
        else:
            # As we output the downstream read, reverse orientation
            if read.is_reverse:
                strand = '-'
            else:
                strand = '+'
            outfile.write("%s\t%i\t%i\t%s\t%i\t%s\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname,
                           read.mapq,
                           strand
                          ))

    c = E.Counter()
    c.input = ninput
    c.removed_insert = nremoved_insert
    c.removed_contig = nremoved_contig
    c.removed_unmapped = nremoved_unmapped
    c.removed_unpaired = nremoved_unpaired
    c.removed_take_only_second = nremoved_take_only_second
    c.removed_coordinate = nremoved_coordinate
    c.output = noutput

    return c



'''
Counts.py - methods for manipulating counts data frames
==========================================================

:Tags: Python

Utility functions for dealing with counts data frames

Requirements:

Code
----

'''
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pandas as pd
from rpy2.robjects import r as R
from rpy2.robjects import pandas2ri
import rpy2.robjects as ro
import numpy as np
import numpy.ma as ma
import copy
import random
import sys
import sklearn.preprocessing as preprocessing

pandas2ri.activate()


def geometric_mean(array, axis=0):
    '''return the geometric mean of an array removing all zero-values but
    retaining total length
    '''
    non_zero = ma.masked_values(array, 0)
    log_a = np.log(non_zero)
    return log_a.mean(axis=axis)
#ma.exp(log_a.mean(axis=axis))


class Counts(object):
    """base class to store counts object

    Attributes
    -----------

    table : pandas DataFrame
       dataframe object with count data

    """

    def __init__(self, filename_or_table):
        # read in table in the constructor for Counts
        # e.g counts = Counts(pd.read_csv(...))

        if isinstance(filename_or_table, str):
            self.table = pd.read_csv(filename_or_table,
                                     sep="\t",
                                     index_col=0)
        else:
            self.table = filename_or_table

        assert self.table.shape, "counts table is empty"

    def clone(self):
        return copy.deepcopy(self)

    def restrict(self, design):
        ''' remove samples not in design '''

        self.table = self.table.ix[:, design.samples]

    def removeSamples(self, min_counts_per_sample=10):
        ''' remove samples without low counts '''

        max_counts_per_sample = self.table.max()

        low_samples = max_counts_per_sample < min_counts_per_sample
        high_samples = [not x for x in low_samples.tolist()]
        sample_names = self.table.columns
        nlow_samples = sum(low_samples)

        if nlow_samples:
            E.warn("%i empty samples are being removed: %s" %
                   (nlow_samples,
                    ",".join([sample_names[x] for x, y in

                              enumerate(low_samples) if y])))
            self.table = self.table.iloc[:, high_samples]

    def removeObservationsFreq(self, min_counts_per_row=1):
        '''remove Observations (e.g genes)

        * remove rows with less than x counts
        '''

        # Remove rows with low counts
        max_counts_per_row = self.table.max(1)
        self.table = self.table[
            max_counts_per_row >= min_counts_per_row]
        observations, samples = self.table.shape
        E.info("trimmed data: %i observations for %i samples" %
               (observations, samples))

    def removeObservationsPerc(self, percentile_rowsums=10):
        '''remove Observations (e.g genes)

        * remove the lowest percentile of rows in the table, sorted
           by total tags per row
        '''

        # percentile filtering
        percentile = float(percentile_rowsums) / 100.0
        sum_counts = self.table.sum(1)
        take = sum_counts >= sum_counts.quantile(percentile)
        E.info("percentile filtering at level %f: keep=%i, discard=%i" %
               (percentile_rowsums,
                sum(take),
                len(take) - sum(take)))
        self.table = self.table[take]

    def normalise(self, method="deseq-size-factors", row_title="total"):
        '''return a table with normalized count data.

        Implemented methods are:

        total-column

           Divide each value by the column total and multiply by 1,000,000

        deseq-size-factors

           Construct a "reference sample" by taking, for each row, the
           geometric mean of the counts in all samples.

           To get the sequencing depth of a column relative to the
           reference, calculate for each row the quotient of the counts in
           your column divided by the counts of the reference sample. Now
           you have, for each row and column, an estimate of the depth
           ratio.

           Simply take the median of all the quotients in a column to get
           the relative depth of the library.

           Divide all values in a column by the normalization factor. This
           normalization method removes all rows with a geometric mean of
           0.

        total-row

           Divide each value in a sample by the value in a particular row.
           The name of the row is given by `row_title`.

        total-count

           Normalise all values in a column by the ratio of the per-column
           sum of counts and the average column count across all rows.

        This method normalises the counts and returns the normalization
        factors that have been applied.

        '''

        if method == "deseq-size-factors":

            # compute row-wise geometric means
            geometric_means = geometric_mean(self.table, axis=1)

            # remove 0 geometric means
            take = geometric_means > 0
            geometric_means = geometric_means[take]
            self.table = self.table[take]

            normed = (self.table.T / geometric_means).T

            self.size_factors = normed.median(axis=0)

            normed = self.table / self.size_factors

        elif method == "edger":

            getCPM = R('''
            suppressMessages(library(edgeR))

            function(counts){
            countsTable = DGEList(counts)
            countsTable = calcNormFactors(countsTable)

            countsTable.cpm <- cpm(countsTable, normalized.lib.sizes=TRUE)

            return (as.data.frame(countsTable.cpm))
            }''' % locals())

            normed = pandas2ri.ri2py(getCPM(pandas2ri.py2ri(self.table)))
            print(normed)

        elif method == "total-count":

            # compute column-wise sum
            column_sums = self.table.sum(axis=0)
            column_sums_mean = np.mean(column_sums)

            self.size_factors = [(column_sums_mean / x) for x in column_sums]
            normed = self.table * self.size_factors

        elif method == "total-column":

            self.size_factors = self.table.sum(axis=0)
            normed = self.table * 1000000.0 / self.size_factors

        elif method == "total-row":

            self.size_factors = self.table.loc[row_title]
            self.table.drop(row_title, inplace=True)
            normed = self.table * 1000000.0 / self.size_factors
        else:
            raise NotImplementedError(
                "normalisation method '%s' not implemented" % method)

        # make sure we did not lose any rows or columns before replacing table
        assert normed.shape == self.table.shape

        self.table = normed

    def sort(self, sort_columns, inplace=True):
        ''' sort counts table by columns'''

        def sort_table(counts):
            counts.table = counts.table.sort(columns=sort_columns)

        if inplace:
            sort_table(self)
            return None

        else:
            tmp_counts = self.clone()
            sort_table(tmp_counts)
            return tmp_counts

    def log(self, base=2, pseudocount=1, inplace=True):
        ''' log transform the counts table '''

        if inplace:
            self.table = np.log(self.table + pseudocount)
            return None

        else:
            tmp_counts = self.clone()
            tmp_counts.table = np.log(tmp_counts.table + pseudocount)
            return tmp_counts

    def transform(self, method="vst", design=None, inplace=True, blind=True):
        '''
        perform transformation on counts table
        current methods are:
         - deseq2 variance stabalising transformation
        - deseq rlog transformation

        Need to supply a design table if not using "blind"
        '''

        assert method in ["vst", "rlog"], ("method must be one of"
                                           "[vst, rlog]")

        method2function = {"vst": "varianceStabilizingTransformation",
                           "rlog": "rlog"}

        t_function = method2function[method]

        r_counts = pandas2ri.py2ri(self.table)

        if not blind:
            assert design, ("if not using blind must supply a design table "
                            "(a CGAT.Expression.ExperimentalDesign object")

            # currently this only accepts "~group" design
            transform = R('''
            function(df, design){

            suppressMessages(library('DESeq2'))

            dds <- suppressMessages(DESeqDataSetFromMatrix(
                     countData= df, colData = design, design = ~group))

            transformed <- suppressMessages(%(t_function)s(dds, blind=FALSE))
            transformed_df <- as.data.frame(assay(transformed))

            return(transformed_df)
            }''' % locals())

            r_design = pandas2ri.py2ri(design.table)
            df = pandas2ri.ri2py(transform(r_counts, r_design))

        else:

            transform = R('''
            function(df){

            suppressMessages(library('DESeq2'))

            design = data.frame(row.names = colnames(df),
                                group = seq(1, length(colnames(df))))

            dds <- suppressMessages(DESeqDataSetFromMatrix(
                     countData= df, colData = design, design = ~group))

            transformed <- suppressMessages(%(t_function)s(dds, blind=TRUE))
            transformed_df <- as.data.frame(assay(transformed))

            return(transformed_df)
            }''' % locals())

            df = pandas2ri.ri2py(transform(r_counts))

        # losing rownames for some reason during the conversion?!
        df.index = self.table.index

        if inplace:
            self.table = df
            # R replaces "-" in column names with ".". Revert back!
            self.table.columns = [x.replace(".", "-")
                                  for x in self.table.columns]
        else:
            tmp_counts = self.clone()
            tmp_counts.table = df
            tmp_counts.table.columns = [x.replace(".", "-")
                                        for x in tmp_counts.table.columns]
            return tmp_counts

    def plotTransformations(self, plot_filename="transformations.png"):
        '''perform transformations and plot to compare'''

        log = self.log(inplace=False)
        vst = self.transform(inplace=False, method="vst")
        rlog = self.transform(inplace=False, method="rlog")

        def counts2meanSdPlot(df):
            ''' takes a counts object and returns a dataframe with standard
            deviation vs. average expression'''

            mean_sd_df = pd.DataFrame({"mean": df.table.apply(np.mean, axis=1),
                                       "std": df.table.apply(np.std, axis=1)})

            mean_sd_df = mean_sd_df[mean_sd_df["mean"] > 0]

            mean_sd_df.sort("mean", inplace=True)
            return mean_sd_df

        plotTransformations = R('''
        function(log_df, rlog_df, vst_df, plot_outfile){

          library(ggplot2)
          library("gridExtra")

          plotMeanSd = function(df, ymax, title){
            df$index <- as.numeric(row.names(df))
            df$bin = .bincode(df$index, seq(1,max(df$index),1000))

            l_txt = element_text(size=20)
            m_txt = element_text(size=15)

            p = ggplot(df, aes(x=index, y=std)) +
                geom_point(alpha=0.1)  +
                stat_summary(mapping=aes(x=bin*1000, y=std),
                             col="red", geom = "point",
                             fun.y = "median", size=5) +
            theme_minimal() +
            ylim(0, ymax) +
            xlab("index") + ylab("sd") +
            theme(axis.text.x=element_text(size=15, angle=90,
                                           hjust=1, vjust=0.5),
                  axis.text.y=m_txt,
                  axis.title.y=m_txt,
                  title=m_txt) +
            ggtitle(title)

            return (p)
        }

        max_sd = max(log_df$std, rlog_df$std, vst_df$std)

        p1 = plotMeanSd(log_df, max_sd, "log")
        p2 = plotMeanSd(rlog_df, max_sd, "rlog")
        p3 = plotMeanSd(vst_df, max_sd, "vst")

        png("%(plot_filename)s",
            width=10, height=10, units="in", res=400, pointsize=1)
        grid.arrange(p1, p2, p3, ncol=3)
        dev.off()
        }''' % locals())

        mean_sd_log_df = counts2meanSdPlot(log)
        mean_sd_vst_df = counts2meanSdPlot(vst)
        mean_sd_rlog_df = counts2meanSdPlot(rlog)

        plotTransformations(mean_sd_log_df, mean_sd_rlog_df, mean_sd_vst_df)

    def zNormalise(self, inplace=True):
        ''' normalise each row to zero mean and unit variance
        (z-score) '''

        samples = self.table.columns
        genes = self.table.index

        z_df = pd.DataFrame(preprocessing.scale(
            self.table, axis=1, with_mean=True, with_std=True, copy=False))

        z_df.index = genes
        z_df.columns = samples

        if inplace:
            self.table = z_df
        else:
            return Counts(z_df)

    def plotDendogram(self, plot_filename=None,
                      distance_method="euclidean",
                      clustering_method="ward.D2"):

        r_counts = pandas2ri.py2ri(self.table)

        makeDendogram = R('''
        function(counts){
          png("%(plot_filename)s")
          par(mar = c(1,4,1,1))
          plot(hclust(dist(t(counts), method = "%(distance_method)s"),
                      method = "%(clustering_method)s"), main="")
          dev.off()
        }''' % locals())

        makeDendogram(r_counts)

    def plotPCA(self, design,
                variance_plot_filename=None, pca_plot_filename=None,
                x_axis="PC1", y_axis="PC2", colour="group", shape="group"):
        ''' use the prcomp function in base R to perform principal components
        analysis.

        Can specify colour and shape as either variables from design table
        or sample names (seperated into id_1, id_2, id_3 based on samples
        having names formated e.g Tissue-Treatment-Replicate)'''

        # TS: swap this for regexes
        assert (x_axis[0:2] == "PC" and y_axis[0:2] == "PC"),\
            "x_axis and y_axis names must start with 'PC'"

        r_counts = pandas2ri.py2ri(self.table)
        r_design = pandas2ri.py2ri(design.table)

        pc_number_1 = int(x_axis.replace("PC", ""))
        pc_number_2 = int(y_axis.replace("PC", ""))

        makePCA = R('''
        function(counts, design){

          suppressMessages(library(ggplot2))
          suppressMessages(library(grid))

          gene_pca <- prcomp(na.omit(t(counts), center = TRUE, scale=TRUE))

          m_text = element_text(size=12)
          s_text = element_text(size=8)


          variance = gene_pca$sdev^2
          variance_explained = round(variance/sum(variance), 5)

          variance_df = data.frame("Variance_explained" = variance_explained,
                                 "PC" = seq(1, length(variance)))
          p_variance = ggplot(variance_df, aes(x=PC, y=Variance_explained))+
          geom_point()+
          geom_line()+
          theme_bw()+
          ylab("Variance explained (%%)")+
          theme(axis.text.x = m_text,
                axis.title.y = m_text,
                axis.title.x = m_text,
                axis.text.y = m_text,
                aspect.ratio=1)

          ggsave("%(variance_plot_filename)s", width=10, height=10, unit="cm")

          PCs_df = data.frame(gene_pca$x)
          PCs_df['sample'] <- rownames(PCs_df)
          design['sample'] <- gsub("-", ".", rownames(design))

          PCs_df = merge(PCs_df, design)

          PCs_df$id_1 = sapply(strsplit(PCs_df$sample, "\\\."), "[", 1)
          PCs_df$id_2 = sapply(strsplit(PCs_df$sample, "\\\."), "[", 2)
          PCs_df$id_3 = sapply(strsplit(PCs_df$sample, "\\\."), "[", 3)

          p_pca = ggplot(PCs_df, aes(x=%(x_axis)s, y=%(y_axis)s)) +
          geom_point(size=3,
                     aes(shape=as.factor(%(shape)s),
                         colour=as.factor(%(colour)s))) +
          scale_colour_discrete(name=guide_legend(title='%(colour)s')) +
          scale_shape_discrete(name=guide_legend(title='%(shape)s')) +
          xlab(paste0('PC%(pc_number_1)i (Variance explained = ' ,
                       round(100 * variance_explained[%(pc_number_1)i], 1),
                       '%%)')) +
          ylab(paste0('PC%(pc_number_2)i (Variance explained = ' ,
                       round(100 * variance_explained[%(pc_number_2)i], 1),
                       '%%)')) +
          theme_bw() +
          theme(axis.text.x = s_text, axis.text.y = s_text,
                title = m_text, legend.text = m_text,
                legend.title = m_text, aspect.ratio=1)

          ggsave("%(pca_plot_filename)s", width=10, height=10, unit="cm")

        }''' % locals())

        makePCA(r_counts, r_design)

    def plotPairwise(self, outfile_scatter, outfile_heatmap, subset=False):
        ''' use the R base pairs function to plot all pairwise
        correlations between the samples

        subset will randomly subset n rows to speed up plotting'''

        plotGGpairs = R('''
        library(ggplot2)
        library(reshape2)

        function(df, df2){

        colnames(df) <- gsub("-", "_", colnames(df))

        width <- height <-  length(colnames(df)) * 100

        png("%(outfile_scatter)s", width=width, height=height, units = "px")

        panel.cor <- function(x, y, digits = 2, prefix = "", cex.cor, ...){
          usr <- par("usr"); on.exit(par(usr))
          par(usr = c(0, 1, 0, 1))
          r <- abs(cor(x, y))
          txt <- format(c(r, 0.123456789), digits = digits)[1]
          txt <- paste0(prefix, txt)
          if(missing(cex.cor)) cex.cor <- 0.8/strwidth(txt)
          text(0.5, 0.5, txt, cex = cex.cor * r * 50)}

        panel.hist = function (x, ...) {
          par(new = TRUE)
          hist(x,
               breaks=30,
               col = "light blue",
               probability = TRUE,
               axes = FALSE,
               main = "")
          rug(x)}

        pairs(df, pch=20, cex=0.1,
              lower.panel = panel.smooth, upper.panel = panel.cor,
              diag.panel=panel.hist)

        dev.off()

        # use the full (not subsetted) df for heatmap correlation
        colnames(df2) <- gsub("-", "_", colnames(df2))
        df_cor <- as.matrix(cor(df2))
        df_cor_melt <- melt(df_cor)

        p = ggplot(df_cor_melt, aes(Var1, Var2, fill=value)) +
        geom_tile() +
        xlab("") + ylab("") +
        scale_fill_continuous(low="white", high="dodgerblue4", name="Correlation") +
        theme_bw() +
        theme(
        axis.text.x = element_text(angle=90, vjust=0.5, hjust=0),
        aspect.ratio=1)

        ggsave("%(outfile_heatmap)s", width=10, height=10, unit="cm")
        }''' % locals())

        if subset:
            if len(self.table.index) > subset:
                rows = random.sample(self.table.index, subset)
                subset_counts = pandas2ri.py2ri(self.table.ix[rows])
            else:
                subset_counts = pandas2ri.py2ri(self.table)
        else:
            subset_counts = pandas2ri.py2ri(self.table)

        plotGGpairs(subset_counts, pandas2ri.py2ri(self.table))

    def heatmap(self, plotfile, zscore=False):
        ''' plots a heatmap
        set zscore=True to use a divergent colour scale
        '''
        # to do: add option to parse design file and add coloured row for
        # variable specified in design file.

        plotHeatmap = R('''
        function(df, zscore){

        library("Biobase")
        library("RColorBrewer")
        library("gplots")

        if(zscore[1]==TRUE){
        #hmcol <- colorRampPalette(colors = c("red", "white", "blue"))
        PuOr <- brewer.pal(11, "PuOr")
        hmcol <- c(colorRampPalette(c(PuOr[1], PuOr[6]))(100),
                   colorRampPalette(c(PuOr[6], PuOr[11]))(100)[-1])
        }
        else{
        hmcol <- colorRampPalette(brewer.pal(9, "GnBu"))(100)
        }

        png("%(plotfile)s", width=1000, height=1000, units="px")

        heatmap.2(as.matrix(df),
                  col = hmcol,
                  scale="none", trace="none", margin=c(18, 10),
                  dendrogram="both", cexCol=2,
                  labRow = "",
                  hclustfun = function(x) hclust(x, method = 'average'),
                  distfun = dist)
        dev.off()
        }''' % locals())

        r_counts = pandas2ri.py2ri(self.table)

        plotHeatmap(r_counts, ro.BoolVector([zscore]))

    def shuffleRows(self,
                    min_cbin, max_cbin, width_cbin,
                    min_ibin, max_ibin, width_ibin,
                    tracks_map,  groups,
                    difference, s_max=100, i=1):
        '''take a dataframe and shuffle the rows to obtain spike in rows.
        return the indices to obtain the rows from the counts table
        and the counts per bin

        The shuffling of rows is parameterised as follows:
        * min_i_bins = minimum bin for initial values
        * width_i_bins = width of bins for initial values
        * max_i_bins = maximum bin for initial values
        * min_c_bins = minimum bin for change values
        * width_c_bins = width of bins for change values
        * max_c_bins = maximum bin for change values
        * tracks_map = dictionary mapping groups to tracks
        * difference = "relative", "logfold" or "abs_logfold"
        * s_max = maximum number of spikes per bin
        * i = number of iterations. More iterations = more filled bins
        '''

        # make bins with an extra bin at the end to capture spike-ins with
        # initial or changes values over the intended range.
        # these are ignored later
        c_bins = np.arange(min_cbin, max_cbin + width_cbin, width_cbin)
        i_bins = np.arange(min_ibin, max_ibin + width_ibin, width_ibin)

        bin_counts = np.zeros((len(i_bins) + 1, len(c_bins) + 1))

        indices = {(key1, key2): []
                   for key1 in np.digitize(i_bins, i_bins)
                   for key2 in np.digitize(c_bins, c_bins)}

        min_occup = min(bin_counts.flatten())
        for iteration in range(0,  i):
            E.info("performing shuffling iteration number %i.." % (
                iteration + 1))
            if min_occup < s_max:
                group1_rand = copy.copy(self.table.index.tolist())
                group2_rand = copy.copy(self.table.index.tolist())
                random.shuffle(group1_rand)
                random.shuffle(group2_rand)

                # subset the dataframe rows in the first random order
                # and by the column_ids in the first group
                # and return means across columns. Repeat for group2
                group1_mean = self.table.ix[group1_rand,
                                            tracks_map[groups[0]]].apply(
                                                np.mean, axis=1).tolist()
                group2_mean = self.table.ix[group2_rand,
                                            tracks_map[groups[1]]].apply(
                                                np.mean, axis=1).tolist()
                # retrieve the index for the bin in
                # which each index value falls
                change_idx, initial_idx = means2idxarrays(
                    group1_mean, group2_mean, i_bins, c_bins, difference)

                # for each initial and change value coordinate
                for idx, coord in enumerate(zip(initial_idx, change_idx)):
                    # ignore spike-in if change or initial fall into the final
                    # bin
                    if coord[0] < len(i_bins) and coord[1] < len(c_bins):
                        if coord in list(indices.keys()):
                            # if max fill of bin not reached
                            if bin_counts[coord] < s_max:
                                # ...append tuple of df indeces for groups
                                bin_counts[coord] += 1
                                indices[coord].append((group1_rand[idx],
                                                       group2_rand[idx]))

                    # find the minimum value in the counts array this should
                    # save time if all bins are filled before final iteration
                    min_occup = min(bin_counts.flatten())

        E.info("The largest bin has %i entries" % max(bin_counts.flatten()))

        return indices, bin_counts

    def outputSpikes(self, indices, tracks_map, groups,
                     output_method, spike_type,
                     min_cbin, width_cbin, max_cbin,
                     min_ibin, width_ibin, max_ibin,
                     min_sbin=1, width_sbin=1, max_sbin=1,
                     append=False):
        ''' method to output spike-ins generated by shuffling rows
        (counts.shuffleRows) or clusters of rows (counts.shuffleCluster)

        parameters:
        * indices - indices for spike-ins
        * min_i_bins = minimum bins for initial values
        * width_i_bins = width of bins for initial values
        * max_i_bins = maximum bin for initial values
        * min_c_bins = minimum bin for change values
        * width_c_bins = width of bins for change values
        * max_c_bins = maximum bin for change values
        * min_s_bins = minimum bin for size values
        * width_s_bins = width of bins for size values
        * max_s_bins = maximum bin for size values
        * tracks_map = dictionary mapping groups to tracks
        * spike_type = "relative" or "logfold"
        * output_method = "append" or "seperate"
        '''

        def makeHeader(tracks_map, groups, keep_columns=None):
            if keep_columns:
                header = keep_columns
            else:
                header = []
            header.extend(tracks_map[groups[0]])
            header.extend(tracks_map[groups[1]])

            return header

        if spike_type == "row":
            index = True
            # keep_columns = ["spike"]
            keep_columns = None

        elif spike_type == "cluster":
            index = False
            keep_columns = ["contig", "position"]

        header = makeHeader(tracks_map, groups, keep_columns=keep_columns)

        if output_method == "append":
            self.table = self.table.ix[:, header]
            self.table.to_csv(sys.stdout, index=index, header=True, sep="\t",
                              dtype={'position': int})
        else:
            sys.stdout.write("%s\t%s\n" % (
                "spike", "\t".join(map(str, header))))

        def getInitialChangeSize(key, width_ibin, min_ibin, width_cbin,
                                 min_cbin, width_s_bin, min_s_bin):
            initial_bin, change_bin, size_bin = key

            # initial and change values are the center of the bin
            initial = ((initial_bin * width_ibin) +
                       min_ibin - (width_ibin * 0.5))
            change = ((change_bin * width_cbin) +
                      min_cbin - (width_cbin * 0.5))
            size = ((size_bin * width_sbin) + min_sbin - 1)
            return initial, change, size

        def getInitialChange(key, width_ibin, min_ibin, width_cbin, min_cbin):
            initial_bin, change_bin = key

            # initial and change values are the center of the bin
            initial = ((initial_bin * width_ibin) +
                       min_ibin - (width_ibin * 0.5))
            change = ((change_bin * width_cbin) +
                      min_cbin - (width_cbin * 0.5))
            return initial, change

        n = 0

        if spike_type == "row":

            for key in indices:
                for pair in indices[key]:
                    initial, change = getInitialChange(
                        key, width_ibin, min_ibin, width_cbin, min_cbin)
                    row = ["_".join(map(str,
                                        ("spike-in", initial, change, n)))]
                    row.extend(self.table.ix[pair[0], tracks_map[groups[0]]])
                    row.extend(self.table.ix[pair[1], tracks_map[groups[1]]])
                    sys.stdout.write("%s\n" % "\t".join(map(str, row)))
                    n += 1

        elif spike_type == "cluster":
            for key in sorted(indices.keys()):
                initial, change, size = getInitialChangeSize(
                    key, width_ibin, min_ibin, width_cbin,
                    min_cbin, width_sbin, min_sbin)
                for values in sorted(indices[key]):
                    (c1s, c1e, c2s, c2e, c1rs, c1re,
                     c2rs, c2re) = values
                    cluster_id = "_".join(
                        map(str, ("spike-in", initial, change,
                                  size, c1rs - c1s, n)))

                    temp_cluster_df = self.table.ix[c1s:c1e, keep_cols]
                    temp_cluster_df['contig'] = cluster_id
                    temp_cluster_swap = self.table.ix[
                        c2rs:c2re, tracks_map[groups[1]]]
                    temp_cluster_swap.set_index(self.table.ix[c1rs:c1re].index,
                                                drop=True,  inplace=True)
                    temp_cluster_df.ix[c1rs:c1re, tracks_map[
                        groups[1]]] = temp_cluster_swap
                    temp_cluster_df.to_csv(sys.stdout, index=index,
                                           header=False, sep="\t",
                                           dtype={'position': int})
                    n += 1


def means2idxarrays(g1, g2, i_bins, c_bins, difference):
    '''take two arrays of values and return the initial values
    and differences as numpy digitised arrays'''

    if difference == "relative":
        # calculate difference between mean values for group1 and group2
        # g1 and g2 always the same length
        change = [g2[x] - g1[x] for x in range(0, len(g1))]
        initial = g1

    elif difference == "logfold":
        change = [np.log2((g2[x] + 1.0) / (g1[x] + 1.0))
                  for x in range(0, len(g1))]
        initial = [np.log2(g1[x] + 1.0) for x in range(0, len(g1))]

    elif difference == "abs_logfold":
        change = [abs(np.log2((g2[x] + 1.0) / (g1[x] + 1.0)))
                  for x in range(0, len(g1))]
        initial = [max(np.log2(g1[x] + 1.0), np.log2(g2[x] + 1.0))
                   for x in range(0, len(g1))]

    # return arrays of len(change) with the index position in c_bins
    # corresponding to the bin in which the value of change falls
    change_idx = np.digitize(change, c_bins, right=True)
    initial_idx = np.digitize(initial, i_bins, right=True)

    return(change_idx, initial_idx)


def findClusters(df, distance, size, tracks_map, groups):
    '''define clusters of genomic loci depending on thresholds for
    size and minimum members per cluster
    This was written with CpGs in mind but will work with any data frame
    containing "position" and "contig" columns'''

    positions = df['position'].tolist()
    contigs = df['contig'].tolist()
    current_pos = 0
    cluster_ix = []
    current_contig = ""
    cluster_dfs = {}
    n = 0
    for ix in range(0, len(positions)):
        next_pos = positions[ix]
        next_contig = contigs[ix]
        if (((next_pos < current_pos + distance) &
             (next_contig == current_contig))):
            cluster_ix.append(ix)
        else:
            if len(cluster_ix) >= size:
                start, end = (cluster_ix[0], cluster_ix[-1] + 1)
                cluster_dfs[n] = df.iloc[start:end]
                n += 1
            cluster_ix = []
            current_pos = next_pos
            current_contig = next_contig

    E.info("found %i clusters" % n)
    return (cluster_dfs)


def shuffleCluster(i_bins, c_bins, tracks_map, groups,
                   difference, s_max, i, clusters_dict,
                   s_bins_max, s_bins_min, s_bins_width):
    '''take a dictionary containing clusters (subdataframes) and shuffle
    subregions of clusters to obtain spike in clusters.
    return indeces from which the spike in clusters can be obtained from the
    original dataframe
    '''
    s_bins = list(range(s_bins_min, s_bins_max + 1, s_bins_width,))

    counts = np.zeros((len(i_bins) + 1, len(c_bins) + 1, len(s_bins) + 1))

    indices = {(key1, key2, key3): []
               for key1 in np.digitize(i_bins, i_bins)
               for key2 in np.digitize(c_bins, c_bins)
               for key3 in np.digitize(s_bins, s_bins)}

    for iteration in range(0,  i):
        E.info("performing shuffling iteration number %i.." % (iteration + 1))
        for size in s_bins:
            group1_mean = []
            group2_mean = []
            g1_rand_s = []
            g2_rand_s = []
            g1_rand = np.random.permutation(list(clusters_dict.keys()))
            g2_rand = np.random.permutation(list(clusters_dict.keys()))
            for perm in range(0, len(g1_rand)):
                cluster1 = clusters_dict[g1_rand[perm]].ix[
                    :, tracks_map[groups[0]]]
                cluster2 = clusters_dict[g2_rand[perm]].ix[
                    :, tracks_map[groups[1]]]
                c1_rand_s = random.randint(
                    min(cluster1.index), max(cluster1.index) - size)
                c1_e = int(c1_rand_s + size)
                c2_rand_s = random.randint(
                    min(cluster2.index), max(cluster2.index) - size)
                c2_e = int(c2_rand_s + size)

                c1_mean = np.mean(np.mean(cluster1.ix[c1_rand_s: c1_e]))
                c2_mean = np.mean(np.mean(cluster2.ix[c2_rand_s: c2_e]))
                group1_mean.append(c1_mean)
                group2_mean.append(c2_mean)
                g1_rand_s.append(c1_rand_s)
                g2_rand_s.append(c2_rand_s)

            change_idx, initial_idx,  = means2idxarrays(
                group1_mean, group2_mean, i_bins,
                c_bins,  difference)
            size_idx = np.digitize([size] * len(initial_idx), s_bins)
            for idx, coord in enumerate(zip(
                    initial_idx, change_idx, size_idx)):
                if counts[coord] < s_max:
                    counts[coord] += 1
                    cluster1_ix = clusters_dict[g1_rand[idx]].index.values
                    cluster1_start = cluster1_ix[0]
                    cluster1_end = cluster1_ix[-1]
                    cluster2_ix = clusters_dict[g2_rand[idx]].index.values
                    cluster2_start = cluster2_ix[0]
                    cluster2_end = cluster2_ix[-1]
                    c1_rand_s = g1_rand_s[idx]
                    c1_rand_e = int(c1_rand_s + size)
                    c2_rand_s = g2_rand_s[idx]
                    c2_rand_e = int(c2_rand_s + size)
                    indices[coord].append((
                        cluster1_start, cluster1_end, cluster2_start,
                        cluster2_end, c1_rand_s, c1_rand_e, c2_rand_s,
                        c2_rand_e))
    return indices, counts


def thresholdBins(indices, counts, s_min):
    '''use counts (np array) to remove bins from indices based on
    threshold (s_min)'''
    output_indices_keep = copy.copy(indices)
    for key in sorted(indices.keys()):
        if counts[key] < s_min:
            output_indices_keep.pop(key)

    E.info("%s/%s bins retained" % (len(list(output_indices_keep.keys())),
                                    len(list(indices.keys()))))
    return output_indices_keep


def loadTagDataPandas(tags_filename, design_filename):
    '''load tag data for deseq/edger analysis.

    *Infile* is a tab-separated file with counts.

    *design_file* is a tab-separated file with the
    experimental design with four columns::

      track   include group   pair
      CW-CD14-R1      0       CD14    1
      CW-CD14-R2      0       CD14    1
      CW-CD14-R3      1       CD14    1
      CW-CD4-R1       1       CD4     1
      FM-CD14-R1      1       CD14    2
      FM-CD4-R2       0       CD4     2
      FM-CD4-R3       0       CD4     2
      FM-CD4-R4       0       CD4     2

    track
        name of track - should correspond to column header in *infile*
    include
        flag to indicate whether or not to include this data
    group
        group indicator - experimental group
    pair
        pair that sample belongs to (for paired tests)

    This method creates various R objects:

    countsTable : data frame with counts.
    groups : vector with groups
    pairs  : vector with pairs

    '''
    counts_table = pd.read_table(tags_filename, sep="\t",
                                 index_col=0, comment="#")

    E.info("read data: %i observations for %i samples" % counts_table.shape)

    E.debug("sample names: %s" % list(counts_table.columns))

    inf = IOTools.openFile(design_filename)
    design_table = pd.read_csv(inf, sep="\t", index_col=0)
    inf.close()

    E.debug("design names: %s" % list(design_table.index))

    missing = set(counts_table.columns).difference(design_table.index)

    if missing:
        E.warn("missing samples from design file are ignored: %s" % missing)

    # remove unnecessary samples
    design_table = design_table[design_table["include"] != 0]
    E.debug("included samples: %s" % list(design_table.index))

    counts_table = counts_table[list(design_table.index)]
    E.info("filtered data: %i observations for %i samples" %
           counts_table.shape)

    return counts_table, design_table


def filterTagDataPandas(counts_table,
                        design_table,
                        filter_min_counts_per_row=1,
                        filter_min_counts_per_sample=10,
                        filter_percentile_rowsums=0):
    '''filter tag data.

    * remove rows with at least x number of counts

    * remove samples with a maximum of *min_sample_counts*

    * remove the lowest percentile of rows in the table, sorted
       by total tags per row
    '''

    # Remove windows with no data
    max_counts_per_row = counts_table.max(1)
    counts_table = counts_table[
        max_counts_per_row >= filter_min_counts_per_row]
    observations, samples = counts_table.shape
    E.info("trimmed data: %i observations for %i samples" %
           (observations, samples))

    # remove samples without data
    max_counts_per_sample = counts_table.max()

    low_samples = max_counts_per_sample < filter_min_counts_per_sample
    high_samples = [not x for x in low_samples.tolist()]
    sample_names = counts_table.columns
    nlow_samples = sum(low_samples)

    if nlow_samples:
        E.warn("%i empty samples are being removed: %s" %
               (nlow_samples,
                ",".join([sample_names[x] for x, y in

                          enumerate(low_samples) if y])))
        counts_table = counts_table.iloc[:, high_samples]

    # percentile filtering
    if filter_percentile_rowsums > 0:
        percentile = float(filter_percentile_rowsums) / 100.0
        sum_counts = counts_table.sum(1)
        take = sum_counts > sum_counts.quantile(percentile)
        E.info("percentile filtering at level %f: keep=%i, discard=%i" %
               (filter_percentile_rowsums,
                sum(take),
                len(take) - sum(take)))
        counts_table = counts_table[take]

    # need to return altered design table based on filtering
    # in case samples are removed!

    design_table = design_table.ix[high_samples]

    nobservations, nsamples = counts_table.shape

    return nobservations, nsamples, counts_table, design_table


def normalizeTagData(counts, method="deseq-size-factors"):
    '''return a table with normalized count data.

    Implemented methods are:

    total-column

       Divide each value by the column total and multiply by 1,000,000

    deseq-size-factors

       Construct a "reference sample" by taking, for each row, the
       geometric mean of the counts in all samples.

       To get the sequencing depth of a column relative to the
       reference, calculate for each row the quotient of the counts in
       your column divided by the counts of the reference sample. Now
       you have, for each row and column, an estimate of the depth
       ratio.

       Simply take the median of all the quotients in a column to get
       the relative depth of the library.

       Divide all values in a column by the normalization factor.

    This method returns the normalized counts and any normalization
    factors that have been applied.

    '''

    if method == "deseq-size-factors":

        # compute row-wise geometric means
        geometric_means = geometric_mean(counts, axis=1)

        # remove 0 geometric means
        take = geometric_means > 0
        geometric_means = geometric_means[take]
        counts = counts[take]

        normed = (counts.T / geometric_means).T

        size_factors = normed.median(axis=0)

        normed = counts / size_factors

    elif method == "total-column":
        size_factors = counts.sum(axis=0)
        normed = counts * 1000000.0 / size_factors
    else:
        raise NotImplementedError(
            "normalization method '%s' not implemented" % method)

    # make sure we did not loose any rows or columns
    assert normed.shape == counts.shape

    return normed, size_factors



'''Experiment.py - Tools for writing reproducible scripts
=========================================================

The :mod:`Experiment` modules contains utility functions for argument
parsing, logging and record keeping within scripts.

This module is imported by most CGAT scripts. It provides convenient
and consistent methods for

   * `Record keeping`_
   * `Argument parsing`_
   * `Input/Output redirection`_
   * `Logging`_
   * `Running external commands`_
   * `Benchmarking`_

See :doc:`../scripts/cgat_script_template` on how to use this module.

The basic usage of this module within a script is::

    """script_name.py - my script

    Mode Documentation
    """
    import sys
    import optparse
    import CGAT.Experiment as E

    def main(argv=None):
        """script main.

        parses command line options in sys.argv, unless *argv* is given.
        """

        if not argv: argv = sys.argv

        # setup command line parser
        parser = E.OptionParser(version="%prog version: $Id$",
                                usage=globals()["__doc__"] )

        parser.add_option("-t", "--test", dest="test", type="string",
                          help="supply help")

        # add common options (-h/--help, ...) and parse
        # command line
        (options, args) = E.Start(parser)

        # do something
        # ...
        E.info("an information message")
        E.warn("a warning message)

        ## write footer and output benchmark information.
        E.Stop()

    if __name__ == "__main__":
        sys.exit(main(sys.argv))

Record keeping
--------------

The central functions in this module are the :py:func:`Start` and
:py:func:`Stop` methods which are called before or after any work is
done within a script.

The :py:func:`Start` is called with an E.OptionParser object.
:py:func:`Start` will add additional command line arguments, such as
``--help`` for command line help or ``--verbose`` to control the
:term:`loglevel`.  It can also add optional arguments for scripts
needing database access, writing to multiple output files, etc.

:py:func:`Start` will write record keeping information to a
logfile. Typically, logging information is output on stdout, prefixed
by a `#`, but it can be re-directed to a separate file. Below is a
typical output::

    # output generated by /ifs/devel/andreas/cgat/beds2beds.py --force-output --exclusive-overlap --method=unmerged-combinations --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz Rela-030m-R1.bed.gz
    # job started at Thu Mar 29 13:06:33 2012 on cgat150.anat.ox.ac.uk -- e1c16e80-03a1-4023-9417-f3e44e33bdcd
    # pid: 16649, system: Linux 2.6.32-220.7.1.el6.x86_64 #1 SMP Fri Feb 10 15:22:22 EST 2012 x86_64
    # exclusive                               : True
    # filename_update                         : None
    # ignore_strand                           : False
    # loglevel                                : 1
    # method                                  : unmerged-combinations
    # output_filename_pattern                 : 030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
    # output_force                            : True
    # pattern_id                              : (.*).bed.gz
    # stderr                                  : <open file \'<stderr>\', mode \'w\' at 0x2ba70e0c2270>
    # stdin                                   : <open file \'<stdin>\', mode \'r\' at 0x2ba70e0c2150>
    # stdlog                                  : <open file \'030m.intersection.tsv.log\', mode \'a\' at 0x1f1a810>
    # stdout                                  : <open file \'<stdout>\', mode \'w\' at 0x2ba70e0c21e0>
    # timeit_file                             : None
    # timeit_header                           : None
    # timeit_name                             : all
    # tracks                                  : None

The header contains information about:

    * the script name (``beds2beds.py``)

    * the command line options (``--force-output --exclusive-overlap
      --method=unmerged-combinations
      --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
      --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz
      Rela-030m-R1.bed.gz``)

    * the time when the job was started (``Thu Mar 29 13:06:33 2012``)
    * the location it was executed (``cgat150.anat.ox.ac.uk``)
    * a unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)
    * the pid of the job (``16649``)

    * the system specification (``Linux 2.6.32-220.7.1.el6.x86_64 #1
      SMP Fri Feb 10 15:22:22 EST 2012 x86_64``)

It is followed by a list of all options that have been set in the script.

Once completed, a script will call the :py:func:`Stop` function to
signify the end of the experiment.

:py:func:`Stop` will output to the log file that the script has
concluded successfully. Below is typical output::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

The footer contains information about:

   * the job has finished (``job finished``)
   * the time it took to execute (``11 seconds``)
   * when it completed (``Thu Mar 29 13:06:44 2012``)
   * some benchmarking information (``11.36  0.45  0.00  0.01``)
     which is ``user time``, ``system time``,
     ``child user time``, ``child system time``.
   * the unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)

The unique job id can be used to easily retrieve matching information
from a concatenation of log files.

Argument parsing
----------------

The module provides :class:`OptionParser` to facilitate option
parsing.  :class:`OptionParser` is derived from the
:py:class:`optparse.OptionParser` class, but has improvements to
provide better formatted output on the command line. It also allows to
provide a comma-separated list to options that accept multiple
arguments. Thus, ``--method=sort --method=crop`` and
``--method=sort,crop`` are equivalent.

Additionally, there are set of commonly used option groups that are
used in many scripts. The :func:`Start` method has options to automatically
add these. For example::

   (options, args) = E.Start(parser, add_output_options=True)

will add the option ``--output-filename-pattern``. Similarly::

   (options, args) = E.Start(parser, add_database_options=True)

will add multiple options for scripts accessing databases, such as
``--database-host`` and ``--database-username``.

Input/Output redirection
------------------------

:func:`Start` adds the options ``--stdin``, ``--stderr` and
``--stdout`` which allow using files as input/output streams.

To make this work, scripts should not read from sys.stdin or write to
sys.stdout directly, but instead use ``options.stdin`` and
``options.stdout``. For example to simply read all lines from stdin
and write to stdout, use::

   (options, args) = E.Start(parser)

   input_data = options.stdin.readlines()
   options.stdout.write("".join(input_data))

The script can then be used in many different contexts::

   cat in.data | python script.py > out.data
   python script.py --stdin=in.data > out.data
   python script.py --stdin=in.data --stdout=out.data

The method handles gzip compressed files transparently. The following
are equivalent::

   zcat in.data.gz | python script.py | gzip > out.data.gz
   python script.py --stdin=in.data.gz --stdout=out.data.gz

For scripts producing multiple output files, use the argument
``add_output_options=True`` to :func:`Start`. This provides the option
``--output-filename-pattern`` on the command line. The user can then
supply a pattern for output files. Any ``%s`` appearing in the pattern
will be substituted by a ``section``. Inside the script, When opening
an output file, use the method :func:`openOutputFile` to provide a
file object::

   output_histogram = E.openOutputFile(section="histogram")
   output_stats = E.openOutputFile(section="stats")

If the user calls the script with::

   python script.py --output-filename-pattern=sample1_%s.tsv.gz

the script will create the files ``sample1_histogram.tsv.gz`` and
``sample1_stats.tsv.gz``.

This method will also add the option ``--force-output`` to permit
overwriting existing files.

Logging
-------

:py:mod:`Experiment` provides the well known logging methods from
the :py:mod:`logging` module such as :py:func:`info`,
:py:func:`warn`, etc. These are provided so that no additional import
of the :py:mod:`logging` module is required, but either functions
can be used.

Frequently, a user wants to know when processing data how many have
been input, how many have been output, how many have been
successfull, etc. The :class:`Counter` class helps in record keeping.
To use, simply create :class:`Counter` object before your main processing
loop, add counts during processing and output a final log message::

   counter = E.Counter()

   for x in range(1000):
      counter.input += 1

      if x < 900:
          counter.output += 1
      else:
          counter.error += 1

   E.log(counter)

:class:`Counter` automatically adds new fields.

Running external commands
-------------------------

The :func:`run` method is a shortcut :py:func:`subprocess.call` and
similar methods with some additional sanity checking.

Benchmarking
------------

The :func:`Start` method records basic benchmarking information when a
script starts and :func:`Stop` outputs it as part of its final log
message::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

See `Record keeping`_ for an explanations of the fields.

To facilitate collecting benchmark information from running multiple
scripts, these data can be tagged and saved in a separate file. See the
command line options ``--timeit``, ``--timeit-name``, ``--timeit-header``
in :func:`Start`.

The module contains some decorator functions for benchmarking
(:func:`benchmark`) and caching function (:func:`cachedfunction`) or
class method (:func:`cachedmethod`) calls.

Complete reference
------------------

'''

import re
import sys
import time
import inspect
import copy
import os
import logging
import collections
import subprocess
import functools
import gzip
import pipes
import optparse
import textwrap
import random
import uuid
import CGAT.IOTools as IOTools


class DefaultOptions:
    stdlog = sys.stdout
    stdout = sys.stdout
    stderr = sys.stderr
    stdin = sys.stdin
    loglevel = 2
    timeit_file = None

global_starting_time = time.time()
global_options = DefaultOptions()
global_args = None
global_id = uuid.uuid4()
global_benchmark = collections.defaultdict(int)

__copyright__ = """
Copyright (c) 2001-2006 Gregory P. Ward.  All rights reserved.
Copyright (c) 2002-2006 Python Software Foundation.  All rights reserved.
Copyright (c) 2011 Yu-Jie Lin.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  * Neither the name of the author nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class BetterFormatter(optparse.IndentedHelpFormatter):
    """A formatter for :class:`OptionParser` outputting indented
    help text.
    """

    def __init__(self, *args, **kwargs):

        optparse.IndentedHelpFormatter.__init__(self, *args, **kwargs)
        self.wrapper = textwrap.TextWrapper(width=self.width)

    def _formatter(self, text):

        return '\n'.join(['\n'.join(p) for p in
                          map(self.wrapper.wrap,
                              self.parser.expand_prog_name(text).split('\n'))])

    def format_description(self, description):

        if description:
            return self._formatter(description) + '\n'
        else:
            return ''

    def format_epilog(self, epilog):

        if epilog:
            return '\n' + self._formatter(epilog) + '\n'
        else:
            return ''

    def format_usage(self, usage):

        return self._formatter(optparse._("Usage: %s\n") % usage)

    def format_option(self, option):
        # Ripped and modified from Python 2.6's optparse's HelpFormatter
        result = []
        opts = self.option_strings[option]
        opt_width = self.help_position - self.current_indent - 2
        if len(opts) > opt_width:
            opts = "%*s%s\n" % (self.current_indent, "", opts)
            indent_first = self.help_position
        else:                       # start help on same line as opts
            opts = "%*s%-*s  " % (self.current_indent, "", opt_width, opts)
            indent_first = 0
        result.append(opts)
        if option.help:
            help_text = self.expand_default(option)
            # Added expand program name
            help_text = self.parser.expand_prog_name(help_text)
            # Modified the generation of help_line
            help_lines = []
            wrapper = textwrap.TextWrapper(width=self.help_width)
            for p in map(wrapper.wrap, help_text.split('\n')):
                if p:
                    help_lines.extend(p)
                else:
                    help_lines.append('')
            # End of modification
            result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            result.extend(["%*s%s\n" % (self.help_position, "", line)
                           for line in help_lines[1:]])
        elif opts[-1] != "\n":
            result.append("\n")
        return "".join(result)


class AppendCommaOption(optparse.Option):

    '''Option with additional parsing capabilities.

    * "," in arguments to options that have the action 'append'
      are treated as a list of options. This is what galaxy does,
      but generally convenient.

    * Option values of "None" and "" are treated as default values.
    '''

    def convert_value(self, opt, value):
        if value is not None:
            if self.nargs == 1:
                if self.action == "append":
                    if "," in value:
                        return [self.check_value(opt, v) for v in
                                value.split(",") if v != ""]
                    else:
                        if value != "":
                            return self.check_value(opt, value)
                        else:
                            return value
                else:
                    return self.check_value(opt, value)
            else:
                return tuple([self.check_value(opt, v) for v in value])

    # why is it necessary to pass action and dest to this function when
    # they could be accessed as self.action and self.dest?
    def take_action(self, action, dest, opt, value, values, parser):

        if action == "append" and type(value) == list:
            values.ensure_value(dest, []).extend(value)
        else:
            optparse.Option.take_action(
                self, action, dest, opt, value, values, parser)


class OptionParser(optparse.OptionParser):

    '''CGAT derivative of OptionParser.
    '''

    def __init__(self, *args, **kwargs):
        # if "--short" is a command line option
        # remove usage from kwargs
        if "--no-usage" in sys.argv:
            kwargs["usage"] = None

        optparse.OptionParser.__init__(self, *args,
                                       option_class=AppendCommaOption,
                                       formatter=BetterFormatter(),
                                       **kwargs)

        # set new option parser
        # parser.formatter = BetterFormatter()
        # parser.formatter.set_parser(parser)
        if "--no-usage" in sys.argv:
            self.add_option("--no-usage", dest="help_no_usage",
                            action="store_true",
                            help="output help without usage information")


class OptionGroup(optparse.OptionGroup):
    pass


def callbackShortHelp(option, opt, value, parser):
    '''output short help (only command line options).'''
    # clear usage and description
    parser.set_description(None)
    parser.set_usage(None)
    # output help
    parser.print_help()
    # exit
    parser.exit()


def openFile(filename, mode="r", create_dir=False):
    '''open file in *filename* with mode *mode*.

    If *create* is set, the directory containing filename
    will be created if it does not exist.

    gzip - compressed files are recognized by the
    suffix ``.gz`` and opened transparently.

    Note that there are differences in the file
    like objects returned, for example in the
    ability to seek.

    returns a file or file-like object.
    '''

    _, ext = os.path.splitext(filename)

    if create_dir:
        dirname = os.path.dirname(filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

    if ext.lower() in (".gz", ".z"):
        return gzip.open(filename, mode)
    else:
        return open(filename, mode)


def getHeader():
    """return a header string with command line options and timestamp

    """
    system, host, release, version, machine = os.uname()

    return "# output generated by %s\n# job started at %s on %s -- %s\n# pid: %i, system: %s %s %s %s" %\
           (" ".join(sys.argv),
            time.asctime(time.localtime(time.time())),
            host,
            global_id,
            os.getpid(),
            system, release, version, machine)


def getParams(options=None):
    """return a string containing script parameters.

    Parameters are all variables that start with ``param_``.
    """
    result = []
    if options:
        members = options.__dict__
        for k, v in sorted(members.items()):
            result.append("# %-40s: %s" % (k, str(v)))
    else:
        vars = inspect.currentframe().f_back.f_locals
        for var in [x for x in list(vars.keys()) if re.match("param_", x)]:
            result.append("# %-40s: %s" %
                          (var, str(vars[var])))

    if result:
        return "\n".join(result)
    else:
        return "# no parameters."


def getFooter():
    """return a header string with command line options and
    timestamp.
    """
    return "# job finished in %i seconds at %s -- %s -- %s" %\
           (time.time() - global_starting_time,
            time.asctime(time.localtime(time.time())),
            " ".join(["%5.2f" % x for x in os.times()[:4]]),
            global_id)


class MultiLineFormatter(logging.Formatter):

    '''logfile formatter: add identation for multi-line entries.'''

    def format(self, record):
        s = logging.Formatter.format(self, record)
        if s.startswith("#"):
            prefix = "#"
        else:
            prefix = ""
        if record.message:
            header, footer = s.split(record.message)
            s = prefix + s.replace('\n', '\n%s' % prefix + ' ' * len(header))
        return s


def Start(parser=None,
          argv=sys.argv,
          quiet=False,
          no_parsing=False,
          add_csv_options=False,
          add_database_options=False,
          add_pipe_options=True,
          add_cluster_options=False,
          add_output_options=False,
          return_parser=False):
    """set up an experiment.

    The :py:func:`Start` method will set up a file logger and add some
    default and some optional options to the command line parser.  It
    will then parse the command line and set up input/output
    redirection and start a timer for benchmarking purposes.

    The default options added by this method are:

    ``-v/--verbose``
        the :term:`loglevel`

    ``timeit``
        turn on benchmarking information and save to file

    ``timeit-name``
         name to use for timing information,

    ``timeit-header``
         output header for timing information.

    ``seed``
         the random seed. If given, the python random
         number generator will be initialized with this
         seed.

    Optional options added are:

    add_csv_options

       ``dialect``
            csv_dialect. the default is ``excel-tab``, defaulting to
            :term:`tsv` formatted files.

    add_database_options
       ``-C/--connection``
           psql connection string
       ``-U/--user``
           psql user name

    add_cluster_options
       ``--use-cluster``
           use cluster
       ``--cluster-priority``
           cluster priority to request
       ``--cluster-queue-manager``
           cluster queue manager to use
       ``--cluster-queue``
           cluster queue to use
       ``--cluster-num-jobs``
           number of jobs to submit to the cluster at the same time
       ``--cluster-memory-resource``
           name of the cluster memory resource (SGE specific option)
       ``--cluster-memory-default``
           default amount of memory allocated per job.
       ``--cluster-options``
           additional options to the cluster for each job.

    add_output_options
       ``-P/--output-filename-pattern``
            Pattern to use for output filenames.

    Arguments
    ---------

    param parser : :py:class:`E.OptionParser`
       instance with command line options.

    argv : list
        command line options to parse. Defaults to
        :py:data:`sys.argv`

    quiet : bool
        set :term:`loglevel` to 0 - no logging

    no_parsing : bool
        do not parse command line options

    return_parser : bool
        return the parser object, no parsing. Useful for inspecting
        the command line options of a script without running it.

    add_csv_options : bool
        add common options for parsing :term:`tsv` separated files.

    add_database_options : bool
        add common options for connecting to various databases.

    add_pipe_options : bool
        add common options for redirecting input/output

    add_cluster_options : bool
        add common options for scripts submitting jobs to the cluster

    add_output_options : bool
        add commond options for working with multiple output files

    Returns
    -------
    tuple
       (:py:class:`E.OptionParser` object, list of positional
       arguments)

    """

    if not parser:
        parser = OptionParser(
            version="%prog version: $Id$")

    global global_options, global_args, global_starting_time

    # save default values given by user
    user_defaults = copy.copy(parser.defaults)

    global_starting_time = time.time()

    group = OptionGroup(parser, "Script timing options")

    group.add_option("--timeit", dest='timeit_file', type="string",
                     help="store timeing information in file [%default].")
    group.add_option("--timeit-name", dest='timeit_name', type="string",
                     help="name in timing file for this class of jobs "
                     "[%default].")
    group.add_option("--timeit-header", dest='timeit_header',
                     action="store_true",
                     help="add header for timing information [%default].")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Common options")

    group.add_option("--random-seed", dest='random_seed', type="int",
                     help="random seed to initialize number generator "
                     "with [%default].")

    group.add_option("-v", "--verbose", dest="loglevel", type="int",
                     help="loglevel [%default]. The higher, the more output.")

    group.add_option("-?", dest="short_help", action="callback",
                     callback=callbackShortHelp,
                     help="output short help (command line options only.")

    parser.add_option_group(group)

    if quiet:
        parser.set_defaults(loglevel=0)
    else:
        parser.set_defaults(loglevel=1)

    parser.set_defaults(
        timeit_file=None,
        timeit_name='all',
        timeit_header=None,
        random_seed=None,
    )

    if add_csv_options:
        parser.add_option("--csv-dialect", dest="csv_dialect", type="string",
                          help="csv dialect to use [%default].")

        parser.set_defaults(
            csv_dialect="excel-tab",
            csv_lineterminator="\n",
        )

    if add_cluster_options:
        group = OptionGroup(parser, "cluster options")
        group.add_option("--no-cluster", "--local", dest="without_cluster",
                         action="store_true",
                         help="do no use cluster - run locally [%default].")
        group.add_option("--cluster-priority", dest="cluster_priority",
                         type="int",
                         help="set job priority on cluster [%default].")
        group.add_option("--cluster-queue-manager", dest="cluster_queue_manager",
                         type="choice",
                         choices=("sge", "slurm"),
                         help="set cluster queue manager [%default].")
        group.add_option("--cluster-queue", dest="cluster_queue",
                         type="string",
                         help="set cluster queue [%default].")
        group.add_option("--cluster-memory-resource", dest="cluster_memory_resource",
                         type="string",
                         help="set cluster memory resource [%default].")
        group.add_option("--cluster-memory-default", dest="cluster_memory_default",
                         type="string",
                         help="set cluster memory default [%default].")
        group.add_option("--cluster-num-jobs", dest="cluster_num_jobs",
                         type="int",
                         help="number of jobs to submit to the queue execute "
                         "in parallel [%default].")
        group.add_option("--cluster-parallel",
                         dest="cluster_parallel_environment",
                         type="string",
                         help="name of the parallel environment to use "
                         "[%default].")
        group.add_option("--cluster-options", dest="cluster_options",
                         type="string",
                         help="additional options for cluster jobs, passed "
                         "on to queuing system [%default].")

        parser.add_option_group(group)

    if add_output_options or add_pipe_options:
        group = OptionGroup(parser, "Input/output options")

        if add_output_options:
            group.add_option(
                "-P", "--output-filename-pattern",
                dest="output_filename_pattern", type="string",
                help="OUTPUT filename pattern for various methods "
                "[%default].")

            group.add_option("-F", "--force-output", dest="output_force",
                             action="store_true",
                             help="force over-writing of existing files.")

            parser.set_defaults(output_filename_pattern="%s",
                                output_force=False)

        if add_pipe_options:

            group.add_option("-I", "--stdin", dest="stdin", type="string",
                             help="file to read stdin from [default = stdin].",
                             metavar="FILE")
            group.add_option("-L", "--log", dest="stdlog", type="string",
                             help="file with logging information "
                             "[default = stdout].",
                             metavar="FILE")
            group.add_option("-E", "--error", dest="stderr", type="string",
                             help="file with error information "
                             "[default = stderr].",
                             metavar="FILE")
            group.add_option("-S", "--stdout", dest="stdout", type="string",
                             help="file where output is to go "
                             "[default = stdout].",
                             metavar="FILE")

            parser.set_defaults(stderr=sys.stderr)
            parser.set_defaults(stdout=sys.stdout)
            parser.set_defaults(stdlog=sys.stdout)
            parser.set_defaults(stdin=sys.stdin)

        parser.add_option_group(group)

    if add_database_options:
        group = OptionGroup(parser, "Database connection options")
        group.add_option(
            "--database-backend", dest="database_backend", type="choice",
            choices=("sqlite", "mysql", "postgres"),
            help="database backend [%default].")
        group.add_option(
            "--database-host", dest="database_host", type="string",
            help="database host [%default].")
        group.add_option(
            "--database-name", dest="database_name", type="string",
            help="name of the database [%default].")
        group.add_option(
            "--database-username", dest="database_username", type="string",
            help="database username [%default].")
        group.add_option(
            "--database-password", dest="database_password", type="string",
            help="database password [%default].")
        group.add_option(
            "--database-port", dest="database_port", type="int",
            help="database port [%default].")

        parser.set_defaults(
            database_backend="sqlite",
            database_name="csvdb",
            database_host="",
            database_port=3306,
            database_username="",
            database_password="")
        parser.add_option_group(group)

    # restore user defaults
    parser.defaults.update(user_defaults)

    if return_parser:
        return parser

    if not no_parsing:
        (global_options, global_args) = parser.parse_args(argv[1:])

    if global_options.random_seed is not None:
        random.seed(global_options.random_seed)

    if add_pipe_options:
        if global_options.stdout != sys.stdout:
            global_options.stdout = openFile(global_options.stdout, "w")
        if global_options.stderr != sys.stderr:
            if global_options.stderr == "stderr":
                global_options.stderr = global_options.stderr
            else:
                global_options.stderr = openFile(global_options.stderr, "w")
        if global_options.stdlog != sys.stdout:
            global_options.stdlog = openFile(global_options.stdlog, "a")
        if global_options.stdin != sys.stdin:
            global_options.stdin = openFile(global_options.stdin, "r")
    else:
        global_options.stderr = sys.stderr
        global_options.stdout = sys.stdout
        global_options.stdlog = sys.stdout
        global_options.stdin = sys.stdin

    if global_options.loglevel >= 1:
        global_options.stdlog.write(getHeader() + "\n")
        global_options.stdlog.write(getParams(global_options) + "\n")
        global_options.stdlog.flush()

    # configure logging
    # map from 0-10 to logging scale
    # 0: quiet
    # 1: little verbositiy
    # >1: increased verbosity
    if global_options.loglevel == 0:
        lvl = logging.ERROR
    elif global_options.loglevel == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    if global_options.stdout == global_options.stdlog:
        format = '# %(asctime)s %(levelname)s %(message)s'
    else:
        format = '%(asctime)s %(levelname)s %(message)s'

    logging.basicConfig(
        level=lvl,
        format=format,
        stream=global_options.stdlog)

    # set up multi-line logging
    # Note that .handlers is not part of the API, might change
    # Solution is to configure handlers explicitely.
    for handler in logging.getLogger().handlers:
        handler.setFormatter(MultiLineFormatter(format))

    return global_options, global_args


def Stop():
    """stop the experiment.

    This method performs final book-keeping, closes the output streams
    and writes the final log messages indicating script completion.
    """

    if global_options.loglevel >= 1 and global_benchmark:
        t = time.time() - global_starting_time
        global_options.stdlog.write(
            "######### Time spent in benchmarked functions #########\n")
        global_options.stdlog.write("# function\tseconds\tpercent\n")
        for key, value in list(global_benchmark.items()):
            global_options.stdlog.write(
                "# %s\t%6i\t%5.2f%%\n" % (key, value,
                                          (100.0 * float(value) / t)))
        global_options.stdlog.write(
            "#######################################################\n")

    if global_options.loglevel >= 1:
        global_options.stdlog.write(getFooter() + "\n")

    # close files
    if global_options.stdout != sys.stdout:
        global_options.stdout.close()
    # do not close log, otherwise error occurs in atext.py
    # if global_options.stdlog != sys.stdout:
    #   global_options.stdlog.close()

    if global_options.stderr != sys.stderr:
        global_options.stderr.close()

    if global_options.timeit_file:

        outfile = open(global_options.timeit_file, "a")

        if global_options.timeit_header:
            outfile.write("\t".join(
                ("name", "wall", "user", "sys", "cuser", "csys",
                 "host", "system", "release", "machine",
                 "start", "end", "path", "cmd")) + "\n")

        csystem, host, release, version, machine = list(map(str, os.uname()))
        uusr, usys, c_usr, c_sys = ["%5.2f" % x for x in os.times()[:4]]
        t_end = time.time()
        c_wall = "%5.2f" % (t_end - global_starting_time)

        if sys.argv[0] == "run.py":
            cmd = global_args[0]
            if len(global_args) > 1:
                cmd += " '" + "' '".join(global_args[1:]) + "'"
        else:
            cmd = sys.argv[0]

        result = "\t".join((global_options.timeit_name,
                            c_wall, uusr, usys, c_usr, c_sys,
                            host, csystem, release, machine,
                            time.asctime(time.localtime(global_starting_time)),
                            time.asctime(time.localtime(t_end)),
                            os.path.abspath(os.getcwd()),
                            cmd)) + "\n"

        outfile.write(result)
        outfile.close()


def log(loglevel, message):
    """log message at loglevel."""
    logging.log(loglevel, message)


def info(message):
    '''log information message, see the :mod:`logging` module'''
    logging.info(message)


def warning(message):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning(message)


def warn(message):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning(message)


def debug(message):
    '''log debugging message, see the :mod:`logging` module'''
    logging.debug(message)


def error(message):
    '''log error message, see the :mod:`logging` module'''
    logging.error(message)


def critical(message):
    '''log critical message, see the :mod:`logging` module'''
    logging.critical(message)


def getOutputFile(section):
    '''return filename to write to, replacing any ``%s`` with section in
    the output pattern for files (``--output-filename-pattern``).
    '''
    return re.sub("%s", section, global_options.output_filename_pattern)


def openOutputFile(section, mode="w"):
    """open file for writing substituting section in the
    output_pattern (if defined).

    If the filename ends with ".gz", the output is opened
    as a gzip'ed file.

    Arguments
    ---------
    section : string
       section will replace any %s in the pattern for output files.

    mode : char
       file opening mode

    Returns
    -------
    File
        an opened file
    """

    fn = getOutputFile(section)
    try:
        if fn == "-":
            return global_options.stdout
        else:
            if not global_options.output_force and os.path.exists(fn):
                raise OSError(
                    ("file %s already exists, use --force-output to "
                     "overwrite existing files.") % fn)
            return IOTools.openFile(fn, mode)
    except AttributeError:
        return global_options.stdout


class Counter(object):
    '''a counter class.

    The counter acts both as a dictionary and a object permitting
    attribute access.

    Counts are automatically initialized to 0.

    Instantiate and use like this::

       c = Counter()
       c.input += 1
       c.output += 2
       c["skipped"] += 1

       print str(c)
    '''

    __slots__ = ["_counts"]

    def __init__(self):
        """Store data returned by function."""
        object.__setattr__(self, "_counts", collections.defaultdict(int))

    def __setitem__(self, key, value):
        self._counts[key] = value

    def __getitem__(self, key):
        return self._counts[key]

    def __getattr__(self, name):
        return self._counts[name]

    def __setattr__(self, name, value):
        self._counts[name] = value

    def __str__(self):
        return ", ".join("%s=%i" % x for x in self._counts.items())

    def __iadd__(self, other):
        try:
            for key, val in other.items():
                self._counts[key] += val
        except:
            raise TypeError("unknown type")
        return self

    def iteritems(self):
        return iter(self._counts.items())

    def items(self):
        return iter(self._counts.items())

    def asTable(self):
        '''return values as tab-separated table (without header).

        Key, value pairs are sorted lexicographically.
        '''
        return '\n'.join("%s\t%i" % x
                         for x in sorted(self._counts.items()))


def run(statement,
        return_stdout=False,
        return_popen=False,
        **kwargs):
    '''execute a command line statement.

    By default this method returns the code returned by the executed
    command. If *return_stdout* is True, the contents of stdout are
    returned as a file object. If *return_popen*, the Popen object is
    returned.

    ``kwargs`` are passed on to subprocess.call,
    subprocess.check_output or subprocess.Popen.

    Raises
    ------

    OSError
       If process failed or was terminated.

    '''

    # remove new lines
    statement = " ".join(re.sub("\t+", " ", statement).split("\n")).strip()

    if "<(" in statement:
        shell = os.environ.get('SHELL', "/bin/bash")
        if "bash" not in shell:
            raise ValueError(
                "require bash for advanced shell syntax: <()")
        # Note: pipes.quote is deprecated. In Py3, use shlex.quote
        # (not present in Py2.7)
        statement = "%s -c %s" % (shell, pipes.quote(statement))

    if return_stdout:
        return subprocess.check_output(statement, shell=True, **kwargs)
    elif return_popen:
        return subprocess.Popen(statement, shell=True, **kwargs)
    else:
        retcode = subprocess.call(statement, shell=True, **kwargs)
        if retcode < 0:
            raise OSError("process was terminated by signal %i" % -retcode)
        return retcode


def benchmark(func):
    """decorator collecting wall clock time spent in decorated method."""

    def wrapper(*arg):
        t1 = time.time()
        res = func(*arg)
        t2 = time.time()
        key = "%s:%i" % (func.__name__, func.__code__.co_firstlineno)
        global_benchmark[key] += t2 - t1
        global_options.stdlog.write(
            '## benchmark: %s completed in %6.4f s\n' % (key, (t2 - t1)))
        global_options.stdlog.flush()
        return res
    return wrapper


def cachedmethod(function):
    '''decorator for caching a method.'''
    return Memoize(function)


class cachedfunction(object):

    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.

    Taken from http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


class Memoize(object):

    def __init__(self, fn):
        self.cache = {}
        self.fn = fn

    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    def __call__(self, *args):
        if args in self.cache:
            return self.cache[args]
        else:
            object = self.cache[args] = self.fn(self.instance, *args)
            return object


def GetFooter():
    return getFooter()


def GetHeader():
    return getHeader()


def GetParams():
    return getParams()



'''
Mali.py - Tools for multiple alignments 
=======================================

:Tags: Python

Code
----

'''

import re
import string
import math
import copy
import random


class AlignedString:

    mGapChars = ("-", ".")
    mGapChar = "-"

    def __init__(self, identifier, fr, to, s, is_mangled=False):
        self.mId = identifier
        self.mFrom = fr
        self.mTo = to
        self.mString = s
        self.mIsMangled = is_mangled

    def __len__(self):
        return len(self.mString)

    def maskColumn(self, column, mask_char="x"):
        self.maskColumns(self, [column, ], mask_char=mask_char)

    def maskColumns(self, columns, mask_char="x"):
        s = list(self.mString)
        for c in columns:
            if s[c] not in self.mGapChars:
                s[c] = mask_char
        self.mString = "".join(s)

    def mapColumns(self, columns, map_function):
        s = list(self.mString)
        for c in columns:
            if s[c] not in self.mGapChars:
                s[c] = map_function(s[c])
        self.mString = "".join(s)

    def takeColumns(self, columns):
        """take columns.

        not implemented yet: updating of residue numbers.
        """
        s = []

        for c in columns:
            s.append(self.mString[c])

        self.mString = "".join(s)

    def truncate(self, first, last):
        """truncate aligned string.

        not implemented yet: updating of residue numbers.
        """
        self.mString = self.mString[first:last]

    def insertColumns(self, position, num_columns, char="-"):
        """insert num_columns columns at position."""
        self.mString = self.mString[:position] + \
            char * num_columns + self.mString[position:]

    def getIdentifier(self):
        '''return demangled identifier.'''
        if self.mIsMangled:
            return self.mId.rsplit("_", 1)[0]
        else:
            return self.mId

    def getResidueNumber(self, position):
        x = self.mFrom
        for c in self.mString[:position]:
            if c not in self.mGapChars:
                x += 1

        return x

    def getSequence(self):
        """return sequence without gaps."""
        return re.sub("[%s]" % "".join(self.mGapChars), "", self.mString)

    def getSegments(self, transitions, gap_chars="-."):
        """return segments of alignment according to
        transitions.

        Transitions is a list of residue numbers.
        """
        segments = []
        pos = self.mFrom

        first_x = 0
        for x in range(len(self.mString)):

            char = self.mString[x]

            if char not in gap_chars:
                if pos in transitions:
                    if first_x != x:
                        segments.append((first_x, x))
                    first_x = x
                pos += 1

        # add last segment, unless it has already been added
        # because it was part of the transitions
        if first_x != x:
            segments.append((first_x, len(self.mString)))
        elif pos in transitions:
            segments.append((first_x, len(self.mString)))

        return segments

    def threadSequence(self, new_sequence, map_old2new):
        """thread a new sequence onto this sequence replacing
        characters in this sequence with those in new sequence.

        ``map_old2new`` lists the positions that should be replaced.
        It maps non-gap positions in old to those in new. Non-gap
        positions than are not mapped are set to :attr:`gapchar`.
        """

        s = list(self.mString)
        c = 0
        for x in range(len(s)):
            m = map_old2new.mapRowToCol(c)
            if m >= 0:
                s[c] = new_sequence[m]
            else:
                s[c] = self.mGapChar
            if s[c] not in self.mGapChars:
                c += 1

        self.mString = "".join(s)


class MaliData:

    def __init__(self, line, gap_chars="-.", mask_chars="Nn"):

        self.mNGaps = len(re.sub("[^%s]" % gap_chars, "", line))
        self.mNMasked = len(re.sub("[^%s]" % mask_chars, "", line))
        self.mNAll = len(line)
        self.mNChars = self.mNAll - self.mNMasked - self.mNGaps

    def __str__(self):
        return "%i\t%i\t%i\t%i" % (self.mNAll, self.mNChars,
                                   self.mNGaps, self.mNMasked)


class Mali:

    mGapChars = ("-", ".")
    mGapPattern = re.compile("[.-]")
    # character used for gaps
    mGapChar = "-"
    # admissable gap characters
    mGapChars = ("-", ".")
    mMaskChar = "X"

    def __init__(self):

        self.mIdentifiers = []
        self.mMali = {}
        self.mLength = 0
        self.mAnnotations = {}
        self.mName = None

        # set to false, if ranges shall not be output.
        self.mWriteRanges = True

    def __contains__(self, key):
        return key in self.mMali

    def __getitem__(self, key):
        return self.mMali[key].mString

    def __len__(self):
        return len(self.mIdentifiers)

    def __delitem__(self, key):
        del self.mMali[key]
        self.mIdentifiers.remove(key)

    def getClone(self):
        return copy.deepcopy(self)

    def iteritems(self):
        return iter(self.mMali.items())

    def items(self):
        return list(self.mMali.items())

    def values(self):
        return list(self.mMali.values())

    def keys(self):
        return list(self.mMali.keys())

    def getName(self):
        return self.mName

    def setName(self, name):
        self.mName = name

    def getIdentifiers(self):
        return self.mIdentifiers

    def getLength(self):
        """deprecated."""
        return self.getNumSequences()

    def getNumSequences(self):
        return self.__len__()

    def getNumColumns(self):
        if self.mIdentifiers:
            return len(self.mMali[self.mIdentifiers[0]])
        else:
            return 0

    def getWidth(self):
        """deprecated."""
        return self.getNumColumns()

    def rename(self, old_name, new_name):
        """rename an entry."""
        if old_name not in self.mMali:
            raise KeyError("%s not in mali" % old_name)
        self.mIdentifiers[self.mIdentifiers.index(old_name)] = new_name
        self.mMali[new_name] = self.mMali[old_name]
        del self.mMali[old_name]

    def isEmpty(self):
        return len(self.mMali) == 0

    def getSequence(self, key):
        return self.mMali[key]

    def getResidueNumber(self, key, position):
        """return residue number in sequence key at position position."""
        return self.mMali[key].getResidueNumber(position)

    def setSequence(self, key, sequence):
        # TODO: updating of residue numbers
        self.mMali[key].mString = sequence

    def getEntry(self, key):
        return self.mMali[key]

    def addEntry(self, s):
        """add an aligned string object."""
        if s.mId in list(self.mMali.keys()):
            raise KeyError("id %s already in mali" % s.mId)

        self.mIdentifiers.append(s.mId)
        self.mMali[s.mId] = s

    def addSequence(self, id, fr, to, sequence):

        if to < 0:
            to = self.countCharacters(sequence)
        s = AlignedString(id, fr, to, sequence)
        self.addEntry(s)

    def countCharacters(self, row):
        return len(row) - len(self.mGapPattern.findall(row))

    def deleteEntry(self, identifier):
        if identifier not in self.mMali:
            raise KeyError("identifier %s not in mali." % identifier)
        del self.mMali[identifier]
        self.mIdentifiers.remove(identifier)

    def getColumns(self):
        """return mali in column orientation."""
        args = [self.mMali[x].mString for x in self.mIdentifiers]
        return ["".join(x) for x in zip(*args)]

    def getConsensus(self, mark_with_gaps=False):
        """return consensus string.

        The consensus string returns the most frequent character per column
        that is not a gap.
        If mark_with_gaps is set to True, positions with any gap characater are
        set to gaps.
        """
        columns = self.getColumns()
        seq = []
        for x in range(len(columns)):
            s = columns[x]
            counts = [(a, s.count(a))
                      for a in set(list(s)).difference(set(self.mGapChar))]
            if mark_with_gaps and self.mGapChar in columns[x]:
                seq.append(self.mGapChar)
            else:
                counts.sort(key=lambda x: -x[1])
                seq.append(counts[0][0])
        return "".join(seq)

    def readFromFile(self, infile, format="fasta"):
        """read multiple alignment from file in various format."""

        self.mMali = {}
        self.mIdentifiers = []

        pattern_parse_ranges = re.compile("(\S+)/(\d+)-(\d+)")

        # read profiles - a profile possibly consists of several
        # entries per file so treat it differently
        if format.lower() == "profile":

            while 1:
                line = infile.readline()
                if not line:
                    return False
                if line[0] != "#":
                    break

            if line[0] != ">":
                raise "expected '>' at as first character in line %s" % line

            try:
                self.mName, length, width = re.match(
                    ">profile=(\S+) length=(\d+) width=(\d+)", line).groups()
            except AttributeError:
                raise "could not parse header line %s" % line
            width = int(width)
            for x in range(0, width):
                id = "seq%i" % x
                self.mIdentifiers.append(id)
                line = infile.readline()
                if not line:
                    raise "expected %i sequences, only got %i" % (width, x)
                self.mMali[id] = AlignedString(
                    id, 0, self.countCharacters(line[:-1]), line[:-1])
            return True

        if isinstance(infile, list) or \
           isinstance(infile, tuple):
            lines = infile
        else:
            lines = infile.readlines()

        if format not in ("stockholm"):
            # save comments
            self.mComments = [x for x in lines if x[0] == "#"]
            lines = [x for x in lines if x[0] != "#"]
        else:
            self.mComments = []

        # remove empty lines
        lines = [x for x in lines if x.strip()]
        if not lines:
            raise AttributeError("empty alignment")

        def getId(id, s):
            x = pattern_parse_ranges.match(id)
            if x:
                id, fr, to = x.groups()
                fr, to = int(fr) - 1, int(to)
            else:
                fr, to = 0, self.countCharacters(s)
                self.mWriteRanges = False

            return id, fr, to

        #######################################################################
        if format.lower() == "plain":

            for line in lines:
                if not line.strip():
                    continue
                data = line[:-1].split("\t")
                id = data[3]
                xid = id
                x = 0
                while xid in self.mMali:
                    xid = id + "-" + str(x)
                    x += 1

                self.addEntry(
                    AlignedString(xid,
                                  int(data[0]) - 1,
                                  int(data[2]),
                                  data[1]))

        #######################################################################
        elif format.lower() == "fasta":
            pattern_identifier = "\S+"
            id = None
            fragments = []
            for line in lines:
                if line[0] == ">":
                    if id:
                        s = re.sub("\s", "", string.join(fragments, ""))
                        id, fr, to = getId(id, s)
                        self.addEntry(AlignedString(id, fr, to, s))

                    id = re.search(
                        "^(%s)" % pattern_identifier, line[1:-1]).group(0)
                    fragments = []
                    continue
                fragments.append(line[:-1])

            s = re.sub("\s", "", string.join(fragments, ""))
            id, fr, to = getId(id, s)
            self.addEntry(AlignedString(id, fr, to, s))

        #######################################################################
        elif format.lower() == "phylip":
            nsequences, nchars = re.split("\s+", lines[0][:-1].strip())
            nsequences = int(nsequences)
            for line in lines[1:]:
                l = line[:-1].strip()
                if not l:
                    continue
                id, sequence = re.match("(\S+)\s+(.*)", l).groups()
                sequence = re.sub("\s", "", sequence)
                if id not in self.mMali:
                    self.mIdentifiers.append(id)
                    self.mMali[id] = []

                self.mMali[id].append(sequence)

            for id, frags in list(self.mMali.items()):
                s = "".join(frags)
                fr, to = 0, self.countCharacters(s)
                self.mMali[id] = AlignedString(id, fr, to, s)

        #######################################################################
        elif format.lower() == "clustal":
            # skip header line
            del lines[0]
            fragments = {}

            # prune lines
            lines = [x.strip() for x in lines]
            # remove empty lines
            lines = [x for x in lines if len(x[:-1]) > 0]

            for line in lines:
                # remove consensus lines
                if line[0] in ("*", ":"):
                    continue

                data = re.split("\s+", line)
                if len(data) != 2:
                    raise ValueError("parsing error in line %s" % line)

                id, fragment = data
                if id not in fragments:
                    fragments[id] = []
                    self.mIdentifiers.append(id)

                fragments[id].append(fragment)

            for id, f in list(fragments.items()):
                s = re.sub("\s", "", string.join(f, ""))
                self.mMali[id] = AlignedString(
                    id, 0, self.countCharacters(s), s)

        elif format.lower() == "stockholm":
            # skip header line
            assert lines[0].startswith(
                "# STOCKHOLM"), "file is not in stockholm format"
            del lines[0]
            fragments = {}
            annotations = {}
            # prune lines
            lines = [x.strip() for x in lines]
            # remove empty lines
            lines = [x for x in lines if len(x[:-1]) > 0]

            for line in lines:
                data = re.split("\s+", line)

                if data[0] == "//":
                    break

                if line[0] == '#':
                    if data[0] == "#=GC":
                        id, fragment = data[1:3]
                    else:
                        self.mComments.append(line)
                        continue
                    if id not in annotations:
                        annotations[id] = []
                    annotations[id].append(fragment)
                else:

                    if len(data) > 2:
                        raise ValueError("parsing error in line %s" % line)
                    elif len(data) == 1:
                        # treat empty alignments/lines
                        id = data[0]
                        fragment = ""
                    else:
                        id, fragment = data

                    if id not in fragments:
                        fragments[id] = []
                        self.mIdentifiers.append(id)

                    fragments[id].append(fragment)

            n = []
            for id in self.mIdentifiers:
                f = fragments[id]

                s = re.sub("\s", "", string.join(f, ""))
                x = pattern_parse_ranges.match(id)
                if x:
                    id, fr, to = x.groups()
                    fr, to = int(fr) - 1, int(to)
                else:
                    fr, to = 0, self.countCharacters(s)

                n.append(id)
                self.mMali[id] = AlignedString(id, fr, to, s)
            self.mIdentifiers = n

            for id, f in list(annotations.items()):
                s = re.sub("\s", "", string.join(f, ""))
                annotations[id] = s
            self.mAnnotations = annotations
        else:
            raise "unknown alignment format %s" % format

        if len(self.mMali) == 0:
            self.mLength = 0
        else:
            self.mLength = min(
                [len(x.mString) for x in list(self.mMali.values())])

    def writeToFile(self, outfile, write_ranges=True, format="plain",
                    options=None):
        """write alignment to file.

        If options is given, these lines are output into the multiple
        alignment.

        """
        if format == "plain-fasta":
            format = "fasta"
            write_ranges = False

        write_ranges = write_ranges and self.mWriteRanges

        if format == "plain":
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                outfile.write("%i\t%s\t%i\t%s\n" % (
                    m.mFrom + 1, m.mString, m.mTo, m.getIdentifier()))

        elif format == "fasta":
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                if write_ranges:
                    outfile.write(
                        ">%s/%i-%i\n%s\n" % (m.getIdentifier(), m.mFrom + 1, m.mTo, m.mString))
                else:
                    outfile.write(">%s\n%s\n" % (identifier, m.mString))

        elif format == "stockholm":
            outfile.write("# STOCKHOLM 1.0\n")

            if options:
                for o in options:
                    outfile.write("%s\n" % o)

            # calculate offset:
            max_l = 0
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                id = m.getIdentifier()
                # tab does not work as separator
                if m.mTo and write_ranges:
                    x = "%s/%i-%i" % (id, m.mFrom + 1, m.mTo)
                else:
                    x = "%s" % (id)
                max_l = max(max_l, len(x))
            for identifier in list(self.mAnnotations.keys()):
                x = "#=GC %s" % identifier
                max_l = max(max_l, len(x))

            format = "%-" + str(max_l) + "s  %s\n"
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                id = m.getIdentifier()
                # tab does not work as separator
                if m.mTo and write_ranges:
                    x = "%s/%i-%i" % (id, m.mFrom + 1, m.mTo)
                else:
                    x = "%s" % (id)

                outfile.write(format % (x, m.mString))

            for identifier, value in list(self.mAnnotations.items()):
                x = "#=GC %s" % identifier
                outfile.write(format % (x, value))

            outfile.write("//\n")

        elif format == "phylip":

            outfile.write("%i %i\n" % (self.getLength(), self.getWidth()))

            for identifier in self.mIdentifiers:
                outfile.write("%s   %s\n" %
                              (identifier, self.mMali[identifier].mString))

        elif format.lower() == "profile":
            if self.mName:
                name = self.mName
            else:
                name = ",".join(self.mIdentifiers)

            outfile.write(">profile=%s length=%i width=%i\n" %
                          (name, self.getWidth(), self.getLength()))

            for identifier in self.mIdentifiers:
                outfile.write("%s\n" % (self.mMali[identifier].mString))

        elif format == "nexus":
            # nexus formatted output - MrBayes conformant.
            outfile.write("#NEXUS\n")
            outfile.write("begin data;\n")
            outfile.write("  dimensions ntax=%i nchar=%i;\n" %
                          (self.getLength(), self.getWidth()))
            outfile.write(
                "  format datatype=dna interleave=no gap=%s;\n" % (self.mGapChar))
            outfile.write("  matrix\n")
            max_len = max([len(x) for x in self.mIdentifiers])
            format = "  %-" + str(max_len) + "s %s\n"
            for identifier in self.mIdentifiers:
                outfile.write(
                    format % (identifier, self.mMali[identifier].mString))
            outfile.write("  ;\n")
            outfile.write("end;\n")

        else:
            raise "unknown alignment format %s" % format

    def removeUnalignedEnds(self):
        """remove unaligned ends in the multiple alignment.

        unaligned ends correspond to lower-case characters.
        """
        pattern_start = re.compile("^([- .a-z]+)")
        pattern_unaligned = re.compile("[a-z]")

        for s in list(self.mMali.values()):

            first = pattern_start.match(s.mString)
            if first:
                first = first.groups()[0]
                nchars = len(pattern_unaligned.findall(first))
                s.mFrom += nchars
                s.mString = self.mGapChar * len(first) + s.mString[len(first):]

            # search from the back end by reversing. This is much faster than
            # using $ from the back.
            last = pattern_start.match(s.mString[::-1])
            if last:
                last = last.groups()[0]
                nchars = len(pattern_unaligned.findall(last))
                s.mTo -= nchars
                l = len(s) - len(last)
                s.mString = s.mString[:l] + self.mGapChar * l

    def upperCase(self):
        """set all characters to upper case."""
        for k, s in list(self.mMali.items()):
            s.mString = s.mString.upper()

    def lowerCase(self):
        """set all characters to lower case."""
        for k, s in list(self.mMali.items()):
            s.mString = s.mString.lower()

    def removeEndGaps(self):
        """remove end gaps.

        end gaps do not include any characters and thus
        the alignment coordinates won't change.
        """

        pattern_start_gaps = re.compile("^([- ]+)")

        min_from = self.mLength
        max_to = 0

        for s in list(self.mMali.values()):

            first = pattern_start_gaps.match(s.mString)
            if first:
                first = first.groups()[0]
                min_from = min(min_from, len(first))

            # search from the back end by reversing. This is much faster than
            # using $ from the back.
            last = pattern_start_gaps.search(s.mString[::-1])
            if last:
                last = last.groups()[0]
                max_to = max(max_to, len(s) - len(last))

        for s in list(self.mMali.values()):
            s.mString = s.mString[min_from:max_to]

        self.mLength = min([x.mString for x in list(self.mMali.values())])

    def insertColumns(self, position, num_gaps, keep_fixed=None, char="-"):
        """insert gaps at position into multiple alignment.

        if keep_constant is a list of identifiers, those are kept constant,
        instead, gaps are added to the end.
        """

        last_pos = min(self.getWidth(), position + num_gaps)
        for id, seq in list(self.mMali.items()):
            if keep_fixed and id in keep_fixed:
                seq.insertColumns(last_pos, num_gaps, char)
            else:
                seq.insertColumns(position, num_gaps, char)

    def removeGaps(self,
                   allowed_gaps=0,
                   minimum_gaps=1,
                   frame=1):
        """remove gappy columns.

        allowed_gaps: number of gaps allowed for column to be kept
        minimum_gaps: number of gaps for column to be removed

        set minimum_gaps to the number of sequences to remove columns
        with all gaps.

        If frame is > 1 (3 most likely), then a whole codon is removed
        as soon as there is one column to be removed.
        """

        self.removePattern(
            match_function=lambda x: x in self.mGapChars,
            allowed_matches=allowed_gaps,
            minimum_matches=minimum_gaps,
            delete_frame=frame)

    def removePattern(self,
                      match_function,
                      allowed_matches=0,
                      minimum_matches=1,
                      delete_frame=1,
                      search_frame=1):
        """remove columns (or group of columns), that match a certain pattern.

        allowed_matches: number of matches allowed so that column is still kept
        minimum_matches: number of matches required for column to be removed

        set minimum_matches to the number of sequences to remove columns
        with all gaps.

        Patterns are matches in search_frame. For example, if frame is
        3, whole codons are supplied to match_function.

        delete_frame specifies the frame for deletion. If it is set to 3,
        codons are removed if already one column matches.

        Example: remove all columns that contain at least one stop-codon:

        removePattern( lambda x: x.upper() in ("TAG", "TAA", "TGA"),
                allowed_matches = 0,
                minimum_matches = 1,
                search_frame = 3,
                delete_frame = 3)

        """
        nmatches = [0] * self.getWidth()

        for s in [x.mString for x in list(self.mMali.values())]:
            for x in range(0, len(s), search_frame):
                segment = s[x:x + search_frame]
                if match_function(segment):
                    nmatches[x] += 1

        columns = []
        delete_columns = []

        for x in range(len(nmatches)):
            if nmatches[x] >= allowed_matches and nmatches[x] < minimum_matches:
                columns.append(x)
            else:
                delete_columns.append(x)

        if delete_frame != 1:
            s = set(columns)
            for x in delete_columns:
                start = int(math.floor(float(x) / delete_frame) * delete_frame)
                end = start + delete_frame
                for c in range(start, end):
                    if c in s:
                        s.remove(c)

            columns = list(s)
            columns.sort()

        self.takeColumns(columns)

    def removeEmptySequences(self):
        """remove sequences that are completely empty.
        """
        new_ids = []
        for id in self.mIdentifiers:
            if self.countCharacters(self.mMali[id].mString) == 0:
                del self.mMali[id]
                continue
            new_ids.append(id)
        self.mIdentifiers = new_ids

    def upper(self):
        """convert all characters in mali to uppercase."""

        for s in list(self.mMali.values()):
            s.mString = s.mString.upper()

    def lower(self):
        """convert all characters in mali to lowercase."""

        for s in list(self.mMali.values()):
            s.mString = s.mString.lower()

    def shiftAlignment(self, map_id2offset):
        """shift alignment by offset."""

        for identifier, m in list(self.mMali.items()):
            if identifier in map_id2offset:
                o = map_id2offset[identifier]
                m.mFrom += o
                m.mTo += o

    def markCodons(self, mode="case"):
        """mark codons.
        """
        for identifier, m in list(self.mMali.items()):
            s = m.mString
            if len(s) % 3 != 0:
                raise "sequence %s not divisible by 3" % (m.mId)

            is_upper = True
            sequence = []
            for x in range(0, len(s), 3):
                if is_upper:
                    sequence.append(s[x:x + 3].upper())
                    is_upper = False
                else:
                    sequence.append(s[x:x + 3].lower())
                    is_upper = True

            m.mString = "".join(sequence)

    def markTransitions(self, map_id2transitions, mode="case"):
        """mark transitions in the multiple alignment.

        if mode == case, then upper/lower case is used for the transitions

        Otherwise, a character given by mode is inserted.
        """

        if mode in ("case", "keep-odd", "keep-even"):

            # check, if the whole alignment needs to be masked/marked:
            if "mali" in map_id2transitions:
                transitions = map_id2transitions["mali"]
                for identifier, s in list(self.mMali.items()):
                    new_chars = []
                    is_upper = True
                    is_first = False

                    for c in range(len(s)):

                        if c in transitions:
                            is_first = True
                            if is_upper:
                                is_upper = False
                            else:
                                is_upper = True

                        x = s.mString[c]

                        if mode == "case":
                            if x in string.lowercase:
                                x = self.mMaskChar
                            if is_upper:
                                x = string.upper(x)
                            else:
                                x = string.lower(x)
                        elif mode == "keep-even":
                            if is_upper:
                                x = self.mGapChar
                        elif mode == "keep-odd":
                            if not is_upper:
                                x = self.mGapChar

                        new_chars.append(x)

                    s.mString = "".join(new_chars)

            # now do individual sequences
            for identifier, s in list(self.mMali.items()):
                if identifier not in map_id2transitions:
                    continue

                new_chars = []
                c = s.mFrom

                is_upper = True
                transitions = map_id2transitions[identifier]
                for x in s.mString:
                    if x in self.mGapChars:
                        pass
                    else:
                        if c in map_id2transitions[identifier]:
                            if is_upper:
                                is_upper = False
                            else:
                                is_upper = True
                        c += 1

                        if x in string.lowercase:
                            x = self.mMaskChar

                        if mode == "case":
                            if is_upper:
                                x = string.upper(x)
                            else:
                                x = string.lower(x)
                        elif mode == "keep-even":
                            if is_upper:
                                x = self.mGapChar
                        elif mode == "keep-odd":
                            if not is_upper:
                                x = self.mGapChar

                    new_chars.append(x)

                s.mString = "".join(new_chars)
        else:
            raise ValueError("character insertion not implemented yet")

    def buildColumnMap(self, other, join_field=None):
        """build map of columns in other to this."""

        if not join_field:
            join_field = other.mIdentifiers[0]

        if join_field not in other.mMali or \
           join_field not in self.mMali:
            raise "line %s not in both alignments." % (join_field)

        this_seq = self.mMali[join_field]
        other_seq = other.mMali[join_field]

        if this_seq.mFrom != other_seq.mFrom or \
           this_seq.mTo != other_seq.mTo:
            raise ValueError("residue ranges for sequence %s doe not correspond." % (
                join_field))

        map_this2other = []

        this_seq = this_seq.mString.upper()
        other_seq = other_seq.mString.upper()

        # position in other
        o = 0
        for c in this_seq:
            if c in self.mGapChars:
                map_this2other.append(None)
            else:
                while other_seq[o] != c:
                    o += 1
                map_this2other.append(o)
                o += 1
        return map_this2other

    def shuffle(self, frame=1):
        """shuffle multiple alignment.

        The frame determines the block size for shuffling. Use 3 for codons in
        a multiple alignment without frame-shifts.
        """
        columns = list(range(self.getNumColumns() // frame))
        random.shuffle(columns)
        if frame > 1:
            cc = []
            for x in columns:
                cc += [x * frame + y for y in range(frame)]
            columns = cc
        self.takeColumns(columns)

    def propagateMasks(self, min_chars=1, mask_char="x"):
        """propagate masked characters to all rows of a multiple alignment
        within a column.

        If there is at least min_chars in a mali column, that are masks,
        propagate the masks to all other rows.
        """

        masks_per_column = {}
        for identifier, s in list(self.mMali.items()):
            r = s.mString.lower()
            for x in range(len(r)):
                if r[x] == mask_char:
                    if x not in masks_per_column:
                        masks_per_column[x] = 0
                    masks_per_column[x] += 1

        columns_to_mask = []
        for c, n in list(masks_per_column.items()):
            if n >= min_chars:
                columns_to_mask.append(c)

        columns_to_mask.sort()

        self.maskColumns(columns_to_mask, mask_char=mask_char)

    def propagateTransitions(self, min_chars=1):
        """propagate lower case in a column to all residues.
        """
        columns_to_change = set()
        for identifier, s in list(self.mMali.items()):
            r = s.mString
            for x in range(len(r)):
                if r[x] in string.lowercase:
                    columns_to_change.add(x)

        columns_to_change = list(columns_to_change)
        columns_to_change.sort()

        self.mapColumns(columns_to_change, string.lower)

    def takeColumns(self, columns):
        """restrict alignments to certain columns."""
        for identifier, s in list(self.mMali.items()):
            s.takeColumns(columns)

        for key, anno in list(self.mAnnotations.items()):
            self.mAnnotations[key] = "".join([anno[c] for c in columns])

    def maskColumns(self, columns, mask_char="x"):
        """mask columns in a multiple alignment."""

        for identifier, s in list(self.mMali.items()):
            s.maskColumns(columns, mask_char=mask_char)

    def mapColumns(self, columns, map_function):
        """apply map_function to all residues in columns."""

        for identifier, s in list(self.mMali.items()):
            s.mapColumns(columns, map_function)

    def recount(self, reset_first=False):
        """recount residue in alignments."""
        for id, seq in list(self.mMali.items()):
            if reset_first:
                seq.mFrom = 0
            seq.mTo = seq.mFrom + self.countCharacters(seq.mString)

    def maskColumn(self, column, mask_char="x"):
        """mask a column."""

        for identifier, s in list(self.mMali.items()):
            s.maskColumn(column, mask_char)

    def copyAnnotations(self, other):
        """copy annotations from annother mali."""

        map_this2other = self.buildColumnMap(other)
        ncols = self.getWidth()

        for key, annotation in list(other.mAnnotations.items()):
            a = []
            for x in range(ncols):
                m = map_this2other[x]
                if m is not None:
                    a.append(annotation[m])
                else:
                    a.append(self.mGapChar)

            self.mAnnotations[key] = "".join(a)

    def getAnnotation(self, key):
        """return annotation associated with key."""
        return self.mAnnotations[key]

    def setAnnotation(self, key, value):
        """set annotation associated with key to value."""
        self.mAnnotations[key] = value

    def addAnnotation(self, key, annotation):
        """add annotation."""
        xkey = key
        n = 0
        while xkey in self.mAnnotations:
            xkey = "%s_%i" % (key, n)

        self.mAnnotations[xkey] = re.sub("\s", "", annotation)

    def truncate(self, first, last):
        """truncate alignment within range."""
        for key, value in list(self.mMali.items()):
            value.truncate(first, last)

    def mapIdentifiers(self, map_old2new=None, pattern_identifier="ID%06i"):
        """map identifiers in multiple aligment.

        if map_old2new is not given, a new map is created (map_new2old)
        """
        new_identifiers = []
        if map_old2new is not None:
            for id in self.mIdentifiers:
                new_id = map_old2new[id]
                new_identifiers.append(new_id)

                entry = self.mMali[id]
                del self.mMali[id]
                # entry.mId = new_id
                self.mMali[new_id] = entry
        else:
            map_old2new = {}
            n = 1
            for id in self.mIdentifiers:
                new_id = pattern_identifier % n
                new_identifiers.append(new_id)

                entry = self.mMali[id]
                del self.mMali[id]
                # entry.mId = new_id
                self.mMali[new_id] = entry

                map_old2new[new_id] = id
                n += 1

        self.mIdentifiers = new_identifiers

        return map_old2new

    def getAlphabet(self):
        """get alphabet from the multiple alignment.

        Alphabet is "na", if more than 90% of characaters are "actgxn",
        otherwise it is "aa".
        """
        s = "".join([x.mString for x in list(self.values())]).lower()
        s = re.sub("[%s]" % "".join(self.mGapChars), "", s)
        ss = re.sub("[acgtxn]", "", s)
        if float(len(ss)) < (len(s) * 0.1):
            return "na"
        else:
            return "aa"

    def checkLength(self):
        """check lengths of aligned strings.

        Return false if they are inconsistent."""

        l = None
        for s in list(self.mMali.values()):
            if l is None:
                l = len(s.mString)
            else:
                if l != len(s.mString):
                    return False
        return True

    def clipByAnnotation(self, key, chars=""):
        """restrict alignment to positions where annotation identified by key
        in chars.

        if chars is empty, nothing is clipped.

        """
        if key not in self.mAnnotations:
            return
        if chars == "":
            return
        if self.getNumColumns() == 0:
            return

        anno = self.mAnnotations[key]
        columns = []

        for x in range(len(anno)):
            if anno[x] in chars:
                columns.append(x)

        self.takeColumns(columns)

    def apply(self, f):
        """apply function f to every row in the multiple alignment."""
        for s in list(self.mMali.values()):
            f(s)

    def filter(self, f):
        '''filter multiple alignment using function *f*.

        The function *f* should return True for entries that should be
        kept and False for those that should be removed.

        '''
        ids, vals = [], []
        for id in self.mIdentifiers:
            s = self.mMali[id]
            if f(s):
                vals.append(s)
                ids.append(id)
        self.mIdentifiers = ids
        self.mMali = vals


class SequenceCollection(Mali):

    """reads in a sequence collection, but permits several entries per id.

    Note that this might cause problems with interleaved formats like
    phylips or clustal.

    This mapping is achieved by appending a numeric suffix. The suffix
    is retained for the life-time of the object, but not output to a
    file.

    """

    def __init__(self):
        Mali.__init__(self)

    def addEntry(self, s):
        """add an aligned string object."""
        id = s.mId
        if id in list(self.mMali.keys()):
            x = 1
            while "%s_%i" % (id, x) in list(self.mMali.keys()):
                x += 1

            id = "%s_%i" % (id, x)
        self.mIdentifiers.append(id)
        s.mIsMangled = True

        self.mMali[id] = s

    def readFromFile(self, infile, format="fasta"):
        """read multiple alignment from file in various format."""

        self.mMali = {}
        self.mIdentifiers = []

        pattern_parse_ranges = re.compile("(\S+)/(\d+)-(\d+)")

        if isinstance(infile, list) or \
           isinstance(infile, tuple):
            lines = infile
        else:
            lines = infile.readlines()

        if format not in ("stockholm"):
            # save comments
            self.mComments = [x for x in lines if x[0] == "#"]
            lines = [x for x in lines if x[0] != "#"]
        else:
            self.mComments = []

        def getId(id, s):
            x = pattern_parse_ranges.match(id)
            if x:
                id, fr, to = x.groups()
                fr, to = int(fr) - 1, int(to)
            else:
                fr, to = 0, self.countCharacters(s)
                self.mWriteRanges = False

            return id, fr, to

        #######################################################################
        if format.lower() == "plain":

            for line in lines:
                if not line.strip():
                    continue
                data = line[:-1].split("\t")
                id = data[3]
                xid = id
                x = 0
                while xid in self.mMali:
                    xid = id + "-" + str(x)
                    x += 1

                self.addEntry(
                    AlignedString(xid,
                                  int(data[0]) - 1,
                                  int(data[2]),
                                  data[1]))

        #######################################################################
        elif format.lower() == "fasta":
            pattern_identifier = "\S+"
            id = None
            fragments = []
            for line in lines:
                if line[0] == ">":
                    if id:
                        s = re.sub("\s", "", string.join(fragments, ""))
                        id, fr, to = getId(id, s)
                        self.addEntry(AlignedString(id, fr, to, s))

                    id = re.search(
                        "^(%s)" % pattern_identifier, line[1:-1]).group(0)
                    fragments = []
                    continue
                fragments.append(line[:-1])

            s = re.sub("\s", "", string.join(fragments, ""))
            id, fr, to = getId(id, s)
            self.addEntry(AlignedString(id, fr, to, s))

        #######################################################################
        elif format.lower() == "phylip":
            raise ValueError("phylip not implemented")

        #######################################################################
        elif format.lower() == "clustal":
            raise ValueError("clustal not implemented")

        elif format.lower() == "stockholm":
            raise ValueError("stockholm not implemented")

        if len(self.mMali) == 0:
            self.mLength = 0
        else:
            self.mLength = min(
                [len(x.mString) for x in list(self.mMali.values())])


def convertMali2Alignlib(mali):
    '''convert a multiple alignment of type :class:`Mali`
    into an alignlib_lite.py_multiple alignment object.
    '''

    import alignlib_lite
    m = alignlib_lite.py_makeMultipleAlignment()
    for identifier in mali.getIdentifiers():
        a = alignlib_lite.py_makeAlignatum(mali[identifier])
        m.add(a)
    return m


def convertAlignlib2Mali(mali, identifiers=None, seqs=None):
    """convert a multiple alignment into an alignlib_lite.py_multiple
    alignment object."""
    m = Mali()

    if not identifiers:
        identifiers = ["%i" % x for x in range(mali.getNumSequences())]

    if seqs is None:
        # old style MultipleAlignment
        for x in range(mali.getNumSequences()):
            a = mali.getRow(x)
            m.addSequence(
                identifiers[x], a.getFrom(), a.getTo(), a.getString())
    else:
        import alignlib_lite
        output = alignlib_lite.py_MultAlignmentFormatPlain(mali, seqs)
        for x in range(mali.getNumSequences()):
            a = output.mData[x]
            m.addSequence(
                identifiers[x], a.getFrom(), a.getTo(), a.getString())

    return m



'''
Biomart.py - Utilities for importing data from biomart
==================================================================

This module provides access to biomart servers using the R biomaRt module.

In order to find out about datasets, filters and attributes, use R to query
the server::

   listDatasets(mart)
   listFilters(mart)
   listAttributes(mart)

Reference
---------

'''

import CGAT.IOTools as IOTools

from rpy2.robjects import r as R
import rpy2.robjects.numpy2ri


def importFromBiomart(outfile,
                      columns,
                      biomart="ensembl",
                      dataset="hsapiens_gene_ensembl",
                      host='www.biomart.org'):
    '''download a dataset from biomart and output as a
    tab-separated table.

    Arguments
    ---------
    outfile : string
       Filename of output file
    columns : dict
       Dictionary mapping biomart columns to columns in the output table.
    biomart : string
       Biomart name
    dataset : string
       Biomart dataset
    host : string
       Biomart host
    '''

    R.library("biomaRt")

    keys = list(columns.keys())

    # The default value for host in the biomaRt package is
    # www.biomart.org but for some reason R errors if you specify
    # host manually but then use the default - but it is fine if
    # host is anything valid apart from www.biomart.org.  So I have
    # changed this to only specify a value if the value you
    # are specifying is different to the default KB

    if host == 'www.biomart.org':
        mart = R.useMart(biomart=biomart, dataset=dataset)
    else:
        mart = R.useMart(biomart=biomart, dataset=dataset, host=host)

    result = R.getBM(attributes=keys, mart=mart)

    outf = IOTools.openFile(outfile, "w")
    outf.write("\t".join([columns[x] for x in keys]) + "\n")

    # for x in ("mim_gene_accession", "mim_morbid_accession"):
    #     result[x] = [ ("", y)[y >= 0] for y in result[x] ]

    for data in zip(*[result[x] for x in keys]):
        outf.write("\t".join(map(str, data)) + "\n")

    outf.close()


def biomart_iterator(columns,
                     biomart="ensembl",
                     dataset="hsapiens_gene_ensembl",
                     host='www.biomart.org',
                     path="/biomart/martservice",
                     filters=None,
                     values=None,
                     archive=False):
    '''download a dataset from biomart and output as a
    tab-separated table.


    Arguments
    ---------
    columns : dict
       List of fields to obtain.
    biomart : string
       Biomart name
    dataset : string
       Biomart dataset
    host : string
       Biomart host
    filters : list
       List of filter to use
    values : list
       Values of the filters
    archive : bool
       If True, use archived version

    Returns
    -------
    iterator
       Iterator over rows in biomart database. Each row
       is dictionary mapping column names to values.

    '''

    R.library("biomaRt")

    # The default value for host in the biomaRt package is
    # www.biomart.org but for some reason R errors if you specify
    # host manually but then use the default - but it is fine if
    # host is anything valid apart from www.biomart.org.  So I have
    # changed this to only specify a value if the value you
    # are specifying is different to the default KB

    if host == 'www.biomart.org':
        mart = R.useMart(biomart=biomart, dataset=dataset, path=path,
                         archive=archive)
    else:
        mart = R.useMart(biomart=biomart, dataset=dataset, host=host,
                         path=path, archive=archive)

    if filters is not None:
        filter_names = rpy2.robjects.vectors.StrVector(list(filters))
    else:
        filter_names = ""

    if values is not None:
        filter_values = values
    else:
        filter_values = ""

    # result is a dataframe
    result = R.getBM(
        attributes=rpy2.robjects.vectors.StrVector(list(columns)),
        filters=filter_names,
        values=filter_values,
        mart=mart)

    # access via result.rx was broken in rpy2 2.4.2, thus try
    # numeric access
    assert tuple(result.colnames) == tuple(columns),\
        "colnames in dataframe: %s different from expected: %s" % \
        (str(tuple(result.colnames)), tuple(columns))

    for data in zip(*[result[x] for x in range(len(columns))]):
        yield dict(list(zip(columns, data)))



"""
Tree.py - A phylogenetic tree
=============================

The :class:`Tree` is derived from the class from Bio.Nexus.Trees.Tree
adding some additional functionality.

Reference
---------

"""
import math

import Bio.Nexus
import Bio.Nexus.Nodes
import Bio.Nexus.Trees


def updateNexus(nexus):
    """change trees in a nexus object (see Biopython_) to :class:`Tree`.
    """
    for x in range(len(nexus.trees)):
        # remove the trailing ";"
        t = nexus.trees[x]

        # in Biopython, to_string includes the name separated by =, while
        # my own class does not.
        tree = t.to_string(branchlengths_only=False, plain=False)
        a = tree.find("=")
        if a >= 0:
            tree = tree[a + 1:]

        nexus.trees[x] = Tree(tree=tree.strip(),
                              weight=t.weight,
                              rooted=t.rooted,
                              name=t.name,
                              data=t.dataclass,
                              max_support=t.max_support)


def Nop(x):
    """empty function for tree traversal"""
    return True


class Tree(Bio.Nexus.Trees.Tree):
    """A phylogenetic tree.

    This class represents a tree using a chain of nodes with on
    predecessor (=ancestor) and multiple successors (=subclades).
    """

    def __init__(self, *args, **kwargs):
        Bio.Nexus.Trees.Tree.__init__(self, *args, **kwargs)

    def __len__(self):
        """returns the number of nodes in the tree."""
        return len(list(self.chain.keys()))

    def root_at_node(self, node, distance=0):
        """root tree at node.

        Arguments
        ---------
        node :
            New root
        distance : float
            Distance of node to new root.

        This is a subset of the code taken from root_with_outgroup.
        """

        def _connect_subtree(parent, child):
            """Hook subtree starting with node child to parent."""
            for i, branch in enumerate(self.unrooted):
                if parent in branch[:2] and child in branch[:2]:
                    branch = self.unrooted.pop(i)
                    break
            else:
                raise ValueError('Unable to connect nodes for rooting: '
                                 'nodes %d and %d are not connected' %
                                 (parent, child))
            self.link(parent, child)
            self.node(child).data.branchlength = branch[2]
            self.node(child).data.support = branch[3]
            # now check if there are more branches connected to the child, and
            # if so, connect them
            child_branches = [b for b in self.unrooted if child in b[:2]]
            for b in child_branches:
                if child == b[0]:
                    succ = b[1]
                else:
                    succ = b[0]
                _connect_subtree(child, succ)

        # code for getting the outgroup node has been removed. Instead
        # outgroup_node is set to node.
        outgroup_node = node

        if outgroup_node == self.root:
            return self.root

        # find ingroup node on a rooted tree
        ingroup_node = self.node(outgroup_node).prev

        # if tree is already rooted with outgroup on a bifurcating root,
        if ingroup_node == self.root and len(self.node(self.root).succ) == 2:
            s = self.node(ingroup_node).succ
            if s[0] == outgroup_node:
                ingroup_node = s[1]
            else:
                ingroup_node = s[0]

        # move to unrooted structure
        self.unroot()

        # now we find the branch that connects outgroup and ingroup
        # print self.node(outgroup_node).prev
        for i, b in enumerate(self.unrooted):
            if outgroup_node in b[:2] and ingroup_node in b[:2]:
                root_branch = self.unrooted.pop(i)
                break
        else:
            raise ValueError('Unrooted and rooted Tree do not match')

        # now we destroy the old tree structure, but keep node data. Nodes will
        # be reconnected according to new outgroup
        for n in self.all_ids():
            self.node(n).prev = None
            self.node(n).succ = []

        # now we just add both subtrees (outgroup and ingroup) branch for
        # branch
        root = Bio.Nexus.Nodes.Node(
            data=Bio.Nexus.Trees.NodeData())            # new root
        self.add(root)                              # add to tree description
        self.root = root.id                           # set as root

        # add branch to ingroup to unrooted tree
        self.unrooted.append(
            [root.id, ingroup_node, root_branch[2] - distance, root_branch[3]])
        # add branch to outgroup to unrooted tree
        self.unrooted.append([root.id, outgroup_node, distance, 0.0])
        _connect_subtree(root.id, ingroup_node)      # add ingroup
        _connect_subtree(root.id, outgroup_node)     # add outgroup

        # if theres still a lonely node in self.chain, then it's the old root,
        # and we delete it
        oldroot = [i for i in self.all_ids() if self.node(
            i).prev is None and i != self.root]
        if len(oldroot) > 1:
            raise ValueError('Isolated nodes in tree description: %s' %
                             ','.join(oldroot))
        elif len(oldroot) == 1:
            self.kill(oldroot[0])
        return self.root

    def to_string(self,
                  support_as_branchlengths=False,
                  branchlengths_only=False,
                  plain=True,
                  write_all_taxa=False,
                  branchlength_format="%1.5f",
                  support_format="%1.2f",
                  format="nexus"):
        """Return a paup compatible tree line.

        Arguments
        ---------
        support_as_branchlengths : bool
           If true, output bootstrap support value as branch lengths.
        branchlengths_only : bool
           Only output branchlengths, no support values
        plain : bool
           Output plain tree (no branch lengths/support values).
        write_all_taxa : bool
           If true, internal node names are output
        branchlength_format : string
           Format to use for branch lengths
        support_format : string
           Format to use for bootstrap support values
        format : string
           Either ``nexus`` on ``NHX``.

        Returns
        -------
        tree : string
            A PAUP compatible tree line.

        """
        # if there's a conflict in the arguments, we override plain=True
        if support_as_branchlengths or branchlengths_only:
            plain = False
        self.support_as_branchlengths = support_as_branchlengths
        self.branchlengths_only = branchlengths_only
        self.plain = plain
        self.write_all_taxa = write_all_taxa

        def make_info_string(data, terminal=False):
            """Creates nicely formatted support/branchlengths."""
            # CHECK FORMATTING

            if format == "nhx":

                info = "[&&NHX"
                try:
                    info += ":S=%s" % data.species
                except AttributeError:
                    pass

                if not terminal:
                    info += ":B=%s" % (support_format % data.support)
                return ':%s%s]' % (branchlength_format % data.branchlength,
                                   info)

            if self.plain:  # plain tree only. That's easy.
                return ''
            # support as branchlengths (eg. PAUP), ignore actual branchlengths
            elif self.support_as_branchlengths:
                if terminal:    # terminal branches have 100% support
                    return ':%1.2f' % (support_format % self.max_support)
                else:
                    return ':%1.2f' % (support_format % data.support)
            # write only branchlengths, ignore support
            elif self.branchlengths_only:
                return ':%s' % (branchlength_format % data.branchlength)
            # write suport and branchlengths (e.g. .con tree of mrbayes)
            else:
                if format in ("nh", "nexus"):
                    if terminal:
                        return ':%s' % (branchlength_format %
                                        data.branchlength)
                    else:
                        # we have blen and suppport
                        if data.branchlength is not None and \
                           data.support is not None:
                            sup = support_format % data.support
                            bl = branchlength_format % data.branchlength
                        # we have only blen
                        elif data.branchlength is not None:
                            sup = support_format % 0
                            bl = branchlength_format % data.branchlength
                        # we have only support
                        elif data.support is not None:
                            sup = support_format % data.support
                            bl = branchlength_format % 0
                        else:
                            sup = support_format % 0
                            bl = branchlength_format % 0
                        return "%s:%s" % (sup, bl)

        def newickize(node):
            """convert a node tree to a newick tree recursively.
            """
            if not self.node(node).succ:  # terminal
                return self.node(node).data.taxon + \
                    make_info_string(self.node(node).data, terminal=True)
            else:
                return '(%s)%s' % (','.join(map(newickize,
                                                self.node(node).succ)),
                                   make_info_string(self.node(node).data))

        def newickize_all_taxa(node):
            if not self.node(node).succ:    # terminal
                return self.node(node).data.taxon + \
                    make_info_string(self.node(node).data, terminal=True)
            else:
                if self.node(node).data.taxon is not None:
                    # changed output: first taxon name, then branch length
                    return '(%s)%s%s' % (
                        ','.join(map(newickize_all_taxa,
                                     self.node(node).succ)),
                        self.node(node).data.taxon,
                        make_info_string(self.node(node).data))
                else:
                    return '(%s)%s' % (
                        ','.join(map(newickize_all_taxa,
                                     self.node(node).succ)),
                        make_info_string(self.node(node).data))

        treeline = ''
        if format == "nexus":
            treeline = 'tree '
            if self.name:
                treeline += self.name
            else:
                treeline += 'a_tree'

            treeline += ' = '
            if self.weight != 1:
                treeline += '[&W%s] ' % str(round(float(self.weight), 3))
            if self.rooted:
                treeline += '[&R] '

        if self.write_all_taxa:
            treeline += '(%s);' % ','.join(map(newickize_all_taxa,
                                               self.node(self.root).succ))
        else:
            treeline += '(%s);' % ','.join(map(newickize,
                                               self.node(self.root).succ))
        return treeline

    def get_nodes(self, node_id):
        """Return a list of nodes downwards from a node (self, node_id).

        The list includes the given node_id.
        """
        if node_id is None:
            node_id = self.root
        if node_id not in self.chain:
            raise ValueError('Unknown node_id: %d.' % node_id)
        list = [node_id]
        for succ in self.chain[node_id].succ:
            list.extend(self.get_nodes(succ))
        return list

    def get_leaves(self, node_id):
        """Return a list of leaf nodes downward from a node (self, node_id).
        """
        return [x for x in self.get_nodes(node_id) if self.node(x).succ == []]

    def root_midpoint(self):
        """perform midpoint rooting of tree.

        The root is placed at equal distance to the two leaves
        furthest apart in the tree (centroid of the tree).
        """

        nnodes = max(self.chain.keys()) + 1

        # distance to root of parent including
        # the branch length of node to parent
        map_N2Root = [0] * (nnodes)
        # distance for node to leaves excluding branch length of this node
        map_N2Leaves = [0] * (nnodes)
        # distance for node to other part of trees excluding its children
        map_N2Other = [0] * (nnodes)

        def dist2leaves(node_id):
            node = self.node(node_id)
            if node.succ:
                map_N2Leaves[node_id] = max(
                    [map_N2Leaves[s] + self.node(s).data.branchlength for s in node.succ])

        def dist2root(node_id):
            node = self.node(node_id)
            map_N2Root[node_id] = node.data.branchlength
            if node.prev is not None:
                map_N2Root[node_id] += map_N2Root[node.prev]

        # traverse tree and record distance to root and maximum distance to
        # leaf at each node
        self.dfs(self.root,
                 pre_function=dist2root,
                 post_function=dist2leaves)

        # update distance to root for each node with length to leaves on the
        # other part of the tree.
        s = set(self.node(self.root).succ)
        for x in s:
            d = max(
                [map_N2Leaves[i] + self.node(i).data.branchlength for i in s.difference(set((x,)))])
            for xx in self.get_nodes(x):
                map_N2Root[xx] += d

        # update distance to any other
        def dist2other(node_id):
            node = self.node(node_id)
            if node.prev is not None:
                p = self.node(node.prev)
                root_dist = map_N2Root[node.prev]
                s = set(p.succ)
                d = max(
                    [map_N2Leaves[i] + self.node(i).data.branchlength for i in s.difference(set((node_id,)))])
                max_d = max(
                    (d, root_dist, map_N2Other[node.prev] + p.data.branchlength))
                map_N2Other[node_id] = max_d

        self.dfs(self.root,
                 pre_function=dist2other)

        # note: need not treat root. Set d2leaves > d2root
        # so that it is never chosen selected.
        map_N2Leaves[self.root] = 1
        map_N2Root[self.root] = 0

        # find node with minimum balance
        best_node = None
        best_balance = None
        for x in range(0, nnodes):
            if x == self.root:
                continue
            node = self.node(x)
            balance = math.fabs(map_N2Other[x] - map_N2Leaves[x])
            # need to add a small quantity to node.data.branchlength
            # to avoid rounding errors for large trees with uniform
            # branchlengths
            if balance >= 0 and balance <= node.data.branchlength + 0.001:
                if best_balance is None or best_balance > balance:
                    best_balance = balance
                    best_node = x

        # in trees with all 0 distances, best_node will be None
        # in this case return a balanced tree
        if best_node is None:
            return self.root_balanced()

        # compute distance of current node to new node
        d1 = map_N2Leaves[best_node]

        distance = float(map_N2Other[
                         best_node] - map_N2Leaves[best_node] + self.node(best_node).data.branchlength) / 2.0

        balance = map_N2Other[best_node] - map_N2Leaves[best_node]
        distance = float(
            balance + self.node(best_node).data.branchlength) / 2.0

        # remove current root
        self.unroot()

        # reroot tree
        result = self.root_at_node(best_node, distance=distance)

    def getNumLeaves(self):
        """return list with number of leaves beyond each node
        """
        nnodes = max(self.chain.keys()) + 1

        counts = [0] * nnodes

        def count(node_id):
            s = self.node(node_id).succ
            if s == []:
                counts[node_id] = 1
            else:
                for n in s:
                    counts[node_id] += counts[n]

        self.dfs(self.root, post_function=count)
        return counts

    def root_balanced(self):
        """perform balanced rooting of tree.

        The root is placed such that the number of leaves
        on either side of the tree is equal (or at most
        different by 1).
        """
        nnodes = max(self.chain.keys()) + 1
        nleaves = len(self.get_terminals())

        # number of leaves
        map_num_leaves = self.getNumLeaves()

        # simply find a node whose number of leaves is nleaves / 2
        threshold = int(math.floor(nleaves / 2.0))
        best_node = None
        for x in range(nnodes):
            if map_num_leaves[x] == threshold:
                best_node = x
                break

        if best_node is None:
            raise "no best node found for %i leaves" % threshold

        # remove current root
        self.unroot()

        # reroot tree
        self.root_at_node(best_node)

    def dfs(self, node_id,
            pre_function=Nop,
            descend_condition=Nop,
            post_function=Nop):
        """dfs tree tree traversal starting at node_id.

        Apply functions pre_function at first and
        post_function at last visit of a node.
        """
        pre_function(node_id)
        for n in self.node(node_id).succ:
            if descend_condition(node_id):
                self.dfs(n, pre_function, descend_condition, post_function)
        post_function(node_id)

    def writeToFile(self, outfile, with_branchlengths=True, format="nh"):
        """write a tree to a file."""

        if format == "nh":
            outfile.write(">%s\n%s\n" % (
                self.name,
                self.to_string(branchlengths_only=with_branchlengths,
                               write_all_taxa=True,
                               format="nh")))

        elif format == "nhx":
            outfile.write(">%s\n%s\n" % (self.name,
                                         self.to_string(format="nhx")))

    def truncate(self, node_id, taxon=None, keep_node=None):
        """truncate tree at node_id.

        This function will not change any branch lengths.
        If keep is given, single child nodes will be collapsed
        until keep_node is reached.
        """
        succ = tuple(self.node(node_id).succ)
        if len(succ) == 0:
            # if this is a leaf, do the same as tree.prune, but
            # nod not change branch length
            prev = self.unlink(node_id)
            self.kill(node_id)
            if not prev == self.root and len(self.node(prev).succ) == 1 \
                    and prev != keep_node:
                succ = self.node(prev).succ[0]
                self.collapse(prev)
        else:
            for s in succ:
                self.unlink(s)
                self.kill(s)

        if taxon:
            self.node(node_id).data.taxon = taxon

    def relabel(self, map_old2new, warn=False):
        """relabel taxa in tree using the provided mapping.
        """
        for node_id, node in list(self.chain.items()):
            if not node.data.taxon:
                continue

            if node.data.taxon in map_old2new:
                node.data.taxon = map_old2new[node.data.taxon]
            else:
                if warn:
                    raise KeyError("taxon %s not in map" % node.data.taxon)

    def rescaleBranchLengths(self, value):
        """rescale branch length so that they sum up to value."""

        t = 0.0
        for node_id, node in list(self.chain.items()):
            t += node.data.branchlength
        if t == 0:
            return
        for node_id, node in list(self.chain.items()):
            node.data.branchlength /= t
        return
